from .config import LANE_TYPE_MAPPING
from .functions import get_coo_list
try:
    from .....Tessng.MyMenu import ProgressDialog as pgd
except:
    class pgd: progress = lambda values, text="": values


def create_network(netiface, links_info, connectors_info, other_info):
    netiface.setNetAttrs("Shapefile Network", otherAttrsJson=other_info)
    print("Network attrs: ", netiface.netAttrs().otherAttrs())
    
    width, height = 600, 400

    error_links = []
    error_connector = []
    
    # 创建路段
    for road_id in pgd.progress(links_info, "创建路段中（7/8）"):
        link_info = links_info[road_id]
        link_type = []
        
        # 获取车道点位
        lanesWithPoints = []
        for lane_id, lane in link_info["lanes_data"].items():
            points = lane["points"]
            lane_type = LANE_TYPE_MAPPING.get(lane["type"], '机动车道')
            link_type.append(lane_type)
            try:
                lanesWithPoints.append({
                    'left': get_coo_list(points['left']),
                    'center': get_coo_list(points['center']),
                    'right': get_coo_list(points['right']),
                })
            except Exception as e:
                error_links.append(f"{road_id}-{lane_id}")
                print(str(e))
                continue
            
            # 场景宽度和高度
            width = max(width, max([abs(p[0]) for p in points["left"]]))
            height = max(height, max([abs(p[1]) for p in points["left"]]))
        
        # 获取路段点位
        # 如果是奇数，取中间车道的中心线
        laneCount = len(lanesWithPoints)
        if laneCount % 2 == 1:
            lCenterLinePoint = lanesWithPoints[int((laneCount-1)/2)]["center"]
        # 如果是偶数，取中间两车道的边线
        else:
            lCenterLinePoint = lanesWithPoints[int(laneCount/2)]["right"]
        
        # 创建路段及车道
        link_obj = netiface.createLink3DWithLanePoints(lCenterLinePoint, lanesWithPoints)
        if link_obj:
            link_obj.setName(f"shp_road_id: {road_id}")
            link_obj.setLaneTypes(link_type)
            links_info[road_id]['obj'] = link_obj
        else:
            error_links.append(f"{road_id}-all")
    
    # 设置场景宽度和高度
    netiface.setSceneSize(width*2+20, height*2+20) # (m)
        
    # 创建连接段
    for connector_road_tuple, connector_info in pgd.progress(connectors_info.items(), "创建连接段中（8/8）"):
        from_link = links_info[connector_road_tuple[0]]['obj']
        to_link = links_info[connector_road_tuple[1]]['obj']
        
        from_link_id, to_link_id = from_link.id(), to_link.id()
        
        from_lane_numbers, to_lane_numbers, lanesWithPoints3 = [], [], []
        for connector in connector_info:
            from_lane_numbers.append(connector['from_lane_number'])
            to_lane_numbers.append(connector['to_lane_number'])
            if connector.get('points'):
                points = connector['points']
                lanesWithPoints3.append(
                    {
                        "center": get_coo_list(points['center']),
                        "left": get_coo_list(points['left']),
                        "right": get_coo_list(points['right']),
                    }
                )

        try:
            if all(lanesWithPoints3):
                netiface.createConnector3DWithPoints(from_link_id, to_link_id, from_lane_numbers, to_lane_numbers,
                                                     lanesWithPoints3, f"{from_link_id}-{to_link_id}")
            else:
                netiface.createConnector(from_link_id, to_link_id, from_lane_numbers, to_lane_numbers)
        except:
            error_connector.append(connector_road_tuple)

    if error_links or error_connector:
        error_message = f"创建失败的路段：{error_links}\n创建失败的连接段：{error_connector}"
        print(error_message)



import numpy as np
from . import config


def p2m(x):
    return x * config.sceneScale

def qtpoint2point(qtpoints):
    x_move = config.x_move
    y_move = config.y_move

    points = []
    for qt_point in qtpoints:
        x = p2m(qt_point.x())
        y = -p2m(qt_point.y())
        z = p2m(qt_point.z())
        x += x_move
        y += y_move
        points.append([x, y, z])
    return points

# 计算向量2相对向量1的旋转角度（-pi~pi）
def clockwise_angle(v1, v2):
    x1, y1 = v1.x, v1.y
    x2, y2 = v2.x, v2.y
    dot = x1 * x2 + y1 * y2
    det = x1 * y2 - y1 * x2
    theta = np.arctan2(det, dot)
    return theta


# 拉取的路段类型
road_class_type = {
    1: ["motorway", "motorway_link"],
    2: [
        "motorway", "motorway_link",
        "trunk", "trunk_link",
        "primary", "primary_link",
        "secondary", "secondary_link",
        "tertiary", "tertiary_link"
    ],
}

# 默认路段类型等级
default_road_class = 3

# 不同道路类型的默认车道数
default_lane_count = {
    "motorway": 3,
    "motorway_link": 3,
    "trunk": 3,
    "trunk_link": 3,
    "primary": 3,
    "primary_link": 3,
    "secondary": 2,
    "secondary_link": 2,
    "tertiary": 2,
    "tertiary_link": 2,
    "other": 1,
    }

# 车道宽度
default_lane_width = 3.5

# osm json数据保存路径
import os
# 这个路径需要按照需求修改
# osm_data_save_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "..", "..", "..", "WorkSpace", "Data", "osm_data")
osm_data_save_path = os.path.join(os.getcwd(), "WorkSpace", "Data", "osm_data")
osm_data_before_process_save_path = os.path.join(osm_data_save_path, "before_process")
osm_data_after_process_save_path = os.path.join(osm_data_save_path, "after_process")



from .functions import get_coo_list
try:
    from .....Tessng.MyMenu import ProgressDialog as pgd
except:
    class pgd: progress = lambda values, text="": values


def create_network(netiface, links_data):
    # 记录属性，如：路网来源
    netiface.setNetAttrs("Excel Network", otherAttrsJson={})
    # 场景尺寸
    width, height = 600, 400

    # 创建路段
    for link_id, link_data in pgd.progress(links_data.items(), '路段创建中（2/2）'):
        linkName = link_data["link_name"]
        linkCount = link_data["link_count"]
        linkPoint = get_coo_list(link_data["link_points"])

        # 创建路段及车道
        try:
            try:
                link_obj = netiface.createLink(linkPoint, linkCount)
            except:
                link_obj = netiface.createLink3D(linkPoint, linkCount)
        except Exception as e:
            link_obj = None
            print(str(e))

        # 判断路段是否存在，并设置属性
        if link_obj:
            link_obj.setName(f"{link_id}-{linkName}")

            # 场景宽度和高度
            width = max(width, max([abs(p[0]) for p in link_data["link_points"]]))
            height = max(height, max([abs(p[1]) for p in link_data["link_points"]]))
    
    # 设置场景宽度和高度
    netiface.setSceneSize(width*2+20, height*2+20) # (m)



from pyproj import Proj
from geopandas import GeoDataFrame
from shapely.geometry import LineString
from . import config
from .functions import p2m, qtpoint2point, point2latlon
try:
    from .....Tessng.MyMenu import ProgressDialog as pgd
except:
    class pgd: progress = lambda values, text="": values


def get_info(netiface, proj_string=""):
    lane_action_type = config.lane_action_type

    # 有投影就转为经纬度，否则就用X/Y
    if proj_string:
        projection = Proj(proj_string)
    else:
        projection = None
    # move
    move_distance = netiface.netAttrs().otherAttrs().get("move_distance")
    if move_distance is None or "tmerc" in proj_string:
        move = {"x_move": 0, "y_move": 0}
    else:
        move = move_distance

    # 创建路段的 GeoDataFrame 保存路段信息
    links = netiface.links()
    lane_features = []
    for link in pgd.progress(links, '路段数据保存中（1/2）'):
        link_id = link.id()
        for lane in link.lanes():
            lane_id = lane.id()
            lane_number = lane.number() + 1
            lane_type = lane_action_type.get(lane.actionType(), "driving")
            lane_width = p2m(lane.width())
            lane_points = qtpoint2point(lane.centerBreakPoint3Ds(), move=move)
            if proj_string:
                lane_points = point2latlon(lane_points, projection)
            feature = {
                'id': lane_id,
                'roadId': link_id,
                'laneNumber': lane_number,
                'type': lane_type,
                'width': lane_width,
                'geometry': LineString(lane_points)
            }
            lane_features.append(feature)
    lane_gdf = GeoDataFrame(lane_features, crs=proj_string)

    # 创建连接段的 GeoDataFrame 保存连接段信息
    connectors = netiface.connectors()
    if connectors:
        connector_features = []
        for connector in pgd.progress(connectors, '连接段数据保存中（2/2）'):
            for lane_connector in connector.laneConnectors():
                from_lane_id = lane_connector.fromLane().id()
                to_lane_id = lane_connector.toLane().id()
                lane_points = qtpoint2point(lane_connector.centerBreakPoint3Ds(), move=move)
                if proj_string:
                    lane_points = point2latlon(lane_points, projection)
                feature = {
                    'preLaneId': from_lane_id,
                    'sucLaneId': to_lane_id,
                    'geometry': LineString(lane_points)
                }
                connector_features.append(feature)
        connector_gdf = GeoDataFrame(connector_features, crs=proj_string)
    else:
        connector_gdf = None
    
    return lane_gdf, connector_gdf



import os
import pandas as pd
try:
    from .....Tessng.MyMenu import ProgressDialog as pgd
except:
    class pgd: progress = lambda values, text="": values


# 获取路段信息
def get_data(file_path):
    # 获取文件后缀
    _, extension = os.path.splitext(file_path)
    # 读取文件
    if extension == ".csv":
        try:
            data = pd.read_csv(file_path, encoding="utf-8")
        except:
            data = pd.read_csv(file_path, encoding="gbk")
    elif extension in [".xlsx", ".xls"]:
        data = pd.read_excel(file_path)
    else:
        raise Exception("Invaild file format !")

    # 保存路段信息
    links_data = {}
    ID = 1
    for col in pgd.progress(data.to_numpy(), '路段数据解析中（1/2）'):
        link_name = col[0] if col[0] else ID
        link_count = int(col[1])
        link_points = [list(map(float, j.split(","))) for j in col[2:] if str(j) != "nan"]
        links_data[ID] = {
            "link_name": link_name,
            "link_count": link_count,
            "link_points": link_points
        }
        ID += 1

    return links_data


# -*- coding: utf-8 -*-

__author__ = "Benjamin Orthen, Stefan Urban"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"

from abc import ABC


class RoadRecord(ABC):
    """Abstract base class to model Records (e.g. ElevationRecord) of the OpenDRIVE
    specification.

    These Records all have attributes start_pos, a, b, c, d.
    The attribute attr which is defined the RoadRecord at a given reference line position
    is calculated with the following equation:
    attr = a + b*ds + c*ds² + d*ds³
    where ds being the distance along the reference line between the start of the entry
    and the actual position.

    ds starts at zero for each RoadRecord.

    The absolute position of an elevation value is calculated by
      s = start_pos + ds



    Attributes:
      start_pos: Position in curve parameter ds where the RoadRecord starts.
      polynomial_coefficients: List of values [a, b, c, d, ...] which can be evaluated with an
        polynomial function.
    """

    def __init__(self, *polynomial_coefficients: float, start_pos: float = None):
        self.start_pos = start_pos
        self.polynomial_coefficients = []
        for coeff in polynomial_coefficients:
            self.polynomial_coefficients.append(coeff)



from shapely.geometry import Polygon
from pyproj import Proj
from . import config
from .functions import p2m, qtpoint2point, calculate_angle_with_y_axis
try:
    from .....Tessng.MyMenu import ProgressDialog as pgd
except:
    class pgd: progress = lambda values, text="": values


def get_data(netiface, proj_string=""):
    data = {
        "header": "",
        "name": "",
        "road": [],
        "connector": [],
        "area": [],
        "crosswalk": [],
    }

    # 如果有经纬度信息
    if proj_string:
        data["header"] = proj_string
        projection = Proj(proj_string)
        isHaveRealLoc = True
    else:
        isHaveRealLoc = False
    # move
    move_distance = netiface.netAttrs().otherAttrs().get("move_distance")
    if move_distance is None or "tmerc" in proj_string:
        move = {"x_move": 0, "y_move": 0}
    else:
        move = move_distance

    # 在遍历连接段的时候记录路段的转向类型
    turnType_mapping = {}
    # 保存面域信息
    areas = netiface.allConnectorArea()
    for area in pgd.progress(areas, '连接段数据保存中（1/2）'):
        area_data = {
            "id": area.id(),
            "incommingRoads": [],
            "outgoingRoads": [],
            "connector": [],
            "crosswalk": []
        }

        # 面域边界点
        area_boundary_points = []
        
        # 保存连接段信息
        for connector in area.allConnector():
            connector_data = {
                'id': connector.id(),
                'areaId': area.id(),
                'predecessor': connector.fromLink().id(),
                'successor': connector.toLink().id(),
                'links': []
            }
            
            area_data["connector"].append(connector_data['id'])
            if connector_data['predecessor'] not in area_data["incommingRoads"]:
                area_data["incommingRoads"].append(connector_data['predecessor'])
            if connector_data['successor'] not in area_data["outgoingRoads"]:
                area_data["outgoingRoads"].append(connector_data['successor'])
            
            # 连接段分车道
            for lane_connector in connector.laneConnectors():
                link_data = {
                    'predecessor': lane_connector.fromLane().id(),
                    'predecessorNumber': lane_connector.fromLane().number(),
                    'successor': lane_connector.toLane().id(),
                    'successorNumber': lane_connector.toLane().number(),
                    'center_points_tess': qtpoint2point(lane_connector.centerBreakPoint3Ds(), "tess", move=move),
                    'left_points_tess': qtpoint2point(lane_connector.leftBreakPoint3Ds(), "tess", move=move),
                    'right_points_tess': qtpoint2point(lane_connector.rightBreakPoint3Ds(), "tess", move=move),
                }
                link_data['start_points_tess'] = link_data['center_points_tess'][0]
                link_data['end_points_tess'] = link_data['center_points_tess'][-1]
                if isHaveRealLoc:
                    link_data['center_points_real'] = qtpoint2point(lane_connector.centerBreakPoint3Ds(), "real", projection, move=move)
                    link_data['left_points_real'] = qtpoint2point(lane_connector.leftBreakPoint3Ds(), "real", projection, move=move)
                    link_data['right_points_real'] = qtpoint2point(lane_connector.rightBreakPoint3Ds(), "real", projection, move=move)
                    link_data['start_points_real'] = link_data['center_points_real'][0]
                    link_data['end_points_real'] = link_data['center_points_real'][-1]
                link_data['length'] = p2m(lane_connector.length())
                
                connector_data['links'].append(link_data)
                
                start_angle = calculate_angle_with_y_axis(link_data['center_points_tess'][0], link_data['center_points_tess'][1])
                end_angle = calculate_angle_with_y_axis(link_data['center_points_tess'][-2], link_data['center_points_tess'][-1])
                angle_diff = (end_angle - start_angle + 180) % 360 - 180
                if -45 < angle_diff < 45:
                    turnType = "直行"
                elif -135 < angle_diff < -45:
                    turnType = "左转"
                elif 45 < angle_diff < 135:
                    turnType = "右转"
                else:
                    turnType = "掉头"

                if link_data['predecessor'] not in turnType_mapping.keys():
                    turnType_mapping[link_data['predecessor']] = []
                turnType_mapping[link_data['predecessor']].append(turnType)
                
                left_points = qtpoint2point(lane_connector.leftBreakPoint3Ds(), "tess", move=move)
                right_points = qtpoint2point(lane_connector.rightBreakPoint3Ds(), "tess", move=move)
                one_lane_points = left_points + right_points[::-1]
                x, y, z = zip(*one_lane_points)
                one_lane_points = list(zip(x, y))
                area_boundary_points.append(one_lane_points)
            
            data['connector'].append(connector_data)
            
        try:
            # 构建多边形对象列表
            polygon_list = [Polygon(coords) for coords in area_boundary_points]
            # 计算多边形的并集
            union_polygon = polygon_list[0]
            for polygon in polygon_list[1:]:
                union_polygon = union_polygon.union(polygon)
            # 提取边界点
            union_boundary_coords = list(union_polygon.exterior.coords)
            
            area_data["points_tess"] = union_boundary_coords
            if isHaveRealLoc:
                area_data["points_real"] = [projection(x, y, inverse=True) for x, y in union_boundary_coords]
        except:
            pass
        
        data["area"].append(area_data)

    # 保存路段信息
    links = netiface.links()
    for link in pgd.progress(links, '路段数据保存中（2/2）'):
        link_data = {}
        link_data["id"] = link.id()
        link_data['points_tess'] = qtpoint2point(link.centerBreakPoint3Ds(), "tess", move=move)
        if isHaveRealLoc:
            link_data['points_real'] = qtpoint2point(link.centerBreakPoint3Ds(), "real", projection, move=move)
        link_data['bearing'] = calculate_angle_with_y_axis(link_data['points_tess'][-2], link_data['points_tess'][-1])
        link_data['lanes'] = []
        # 路段分车道
        for lane in link.lanes():
            lane_data = {
                'id': lane.id(),
                'type': lane.actionType(),
                'center_points_tess': qtpoint2point(lane.centerBreakPoint3Ds(), "tess", move=move),
                'left_points_tess': qtpoint2point(lane.leftBreakPoint3Ds(), "tess", move=move),
                'right_points_tess': qtpoint2point(lane.rightBreakPoint3Ds(), "tess", move=move)
            }
            lane_data['start_points_tess'] = lane_data['center_points_tess'][0]
            lane_data['end_points_tess'] = lane_data['center_points_tess'][-1]

            if isHaveRealLoc:
                lane_data['center_points_real'] = qtpoint2point(lane.centerBreakPoint3Ds(), "real", projection, move=move)
                lane_data['left_points_real'] = qtpoint2point(lane.leftBreakPoint3Ds(), "real", projection, move=move)
                lane_data['right_points_real'] = qtpoint2point(lane.rightBreakPoint3Ds(), "real", projection, move=move)
                lane_data['start_points_real'] = lane_data['center_points_real'][0]
                lane_data['end_points_real'] = lane_data['center_points_real'][-1]
            lane_data['length'] = p2m(lane.length())
            lane_data['laneNumber'] = lane.number()
            lane_data['turnType'] = list(set(turnType_mapping.get(lane.id(), [])))
            lane_data['limitSpeed'] = p2m(link.limitSpeed()) # km/h
            
            link_data['lanes'].append(lane_data)
        
        data['road'].append(link_data)
    
    return data



# -*- coding: utf-8 -*-

from .road_record import RoadRecord


class Lanes:
    """ """

    def __init__(self):
        self._laneOffsets = []
        self._lane_sections = []

    @property
    def laneOffsets(self):
        """ """
        self._laneOffsets.sort(key=lambda x: x.start_pos)
        return self._laneOffsets

    @property
    def lane_sections(self):
        """ """
        self._lane_sections.sort(key=lambda x: x.sPos)
        return self._lane_sections

    def getLaneSection(self, laneSectionIdx):
        """

        Args:
          laneSectionIdx:

        Returns:

        """
        for laneSection in self.lane_sections:
            if laneSection.idx == laneSectionIdx:
                return laneSection

        return None

    def getLastLaneSectionIdx(self):
        """ """

        numLaneSections = len(self.lane_sections)

        if numLaneSections > 1:
            return numLaneSections - 1

        return 0


class LaneOffset(RoadRecord):
    """The lane offset record defines a lateral shift of the lane reference line
    (which is usually identical to the road reference line).

    (Section 5.3.7.1 of OpenDRIVE 1.4)

    """


class LeftLanes:
    """ """

    sort_direction = False

    def __init__(self):
        self._lanes = []

    @property
    def lanes(self):
        """ """
        self._lanes.sort(key=lambda x: x.id, reverse=self.sort_direction)
        return self._lanes


class CenterLanes(LeftLanes):
    """ """


class RightLanes(LeftLanes):
    """ """

    sort_direction = True


class Lane:
    """ """

    laneTypes = [
        "none",
        "driving",
        "stop",
        "shoulder",
        "biking",
        "sidewalk",
        "border",
        "restricted",
        "parking",
        "bidirectional",
        "median",
        "special1",
        "special2",
        "special3",
        "roadWorks",
        "tram",
        "rail",
        "entry",
        "exit",
        "offRamp",
        "onRamp",
        "curb",
        "connectingRamp",
    ]

    def __init__(self, parentRoad, lane_section):
        self._parent_road = parentRoad
        self._id = None
        self._type = None
        self._level = None
        self._link = LaneLink()
        self._widths = []
        self._borders = []
        self._road_marks = []
        self.lane_section = lane_section
        self.has_border_record = False

    @property
    def parentRoad(self):
        """ """
        return self._parent_road

    @property
    def id(self):
        """ """
        return self._id

    @id.setter
    def id(self, value):
        self._id = int(value)

    @property
    def type(self):
        """ """
        return self._type

    @type.setter
    def type(self, value):
        if value not in self.laneTypes:
            raise Exception()

        self._type = str(value)

    @property
    def level(self):
        """ """
        return self._level

    @level.setter
    def level(self, value):
        if value not in ["true", "false"] and value is not None:
            raise AttributeError("Value must be true or false.")

        self._level = value == "true"

    @property
    def link(self):
        """ """
        return self._link

    @property
    def widths(self):
        """ """
        self._widths.sort(key=lambda x: x.start_offset)
        return self._widths

    @widths.setter
    def widths(self, value):
        """"""
        self._widths = value

    def getWidth(self, widthIdx):
        """

        Args:
          widthIdx:

        Returns:

        """
        for width in self._widths:
            if width.idx == widthIdx:
                return width

        return None

    def getLastLaneWidthIdx(self):
        """Returns the index of the last width sector of the lane"""

        numWidths = len(self._widths)

        if numWidths > 1:
            return numWidths - 1

        return 0

    @property
    def borders(self):
        """ """
        return self._borders

    @property
    def road_marks(self):
        """ """
        return self._road_marks


class LaneLink:
    """ """

    def __init__(self):
        self._predecessor = None
        self._successor = None

    @property
    def predecessorId(self):
        """ """
        return self._predecessor

    @predecessorId.setter
    def predecessorId(self, value):
        self._predecessor = int(value)

    @property
    def successorId(self):
        """ """
        return self._successor

    @successorId.setter
    def successorId(self, value):
        self._successor = int(value)


class LaneSection:
    """The lane section record defines the characteristics of a road cross-section.

    (Section 5.3.7.2 of OpenDRIVE 1.4)

    """

    def __init__(self, road=None):
        self.idx = None
        self.sPos = None
        self._singleSide = None
        self._leftLanes = LeftLanes()
        self._centerLanes = CenterLanes()
        self._rightLanes = RightLanes()

        self._parentRoad = road

    @property
    def singleSide(self):
        """Indicator if lane section entry is valid for one side only."""
        return self._singleSide

    @singleSide.setter
    def singleSide(self, value):
        if value not in ["true", "false"] and value is not None:
            raise AttributeError("Value must be true or false.")

        self._singleSide = value == "true"

    @property
    def leftLanes(self):
        """Get list of sorted lanes always starting in the middle (lane id -1)"""
        return self._leftLanes.lanes

    @property
    def centerLanes(self):
        """ """
        return self._centerLanes.lanes

    @property
    def rightLanes(self):
        """Get list of sorted lanes always starting in the middle (lane id 1)"""
        return self._rightLanes.lanes

    @property
    def allLanes(self):
        """Attention! lanes are not sorted by id"""
        return self._leftLanes.lanes + self._centerLanes.lanes + self._rightLanes.lanes

    def getLane(self, lane_id: int) -> Lane:
        """

        Args:
          lane_id:

        Returns:

        """
        for lane in self.allLanes:
            if lane.id == lane_id:
                return lane

        return None

    @property
    def parentRoad(self):
        """ """
        return self._parentRoad


class LaneWidth(RoadRecord):
    """Entry for a lane describing the width for a given position.
    (Section 5.3.7.2.1.2.0 of OpenDRIVE 1.4)


    start_offset being the offset of the entry relative to the preceding lane section record

"""

    def __init__(
            self,
            *polynomial_coefficients: float,
            idx: int = None,
            start_offset: float = None
    ):
        self.idx = idx
        self.length = 0
        super().__init__(*polynomial_coefficients, start_pos=start_offset)

    @property
    def start_offset(self):
        """Return start_offset, which is the offset of the entry to the
        start of the lane section.
        """
        return self.start_pos

    @start_offset.setter
    def start_offset(self, value):
        self.start_pos = value


class LaneBorder(LaneWidth):
    """Describe lane by width in respect to reference path.

    (Section 5.3.7.2.1.2.0 of OpenDRIVE 1.4)

    Instead of describing lanes by their width entries and, thus,
    invariably depending on influences of inner
    lanes on outer lanes, it might be more convenient to just describe
    the outer border of each lane
    independent of any inner lanes’ parameters.
    """



from xml.dom import minidom
from . import config
from .node import Doc
from .models import Junction, Connector, Road
try:
    from .....Tessng.MyMenu import ProgressDialog as pgd
except:
    class pgd: progress = lambda values, text="": values


def get_data(netiface, proj_string:str):
    # move
    move = netiface.netAttrs().otherAttrs().get("move_distance")
    if not move or "tmerc" in proj_string:
        config.x_move = 0
        config.y_move = 0
    else:
        config.x_move = -move["x_move"]
        config.y_move = -move["y_move"]

    # 遍历连接段
    connectors = []
    junctions = []
    areas = netiface.allConnectorArea()
    for ConnectorArea in pgd.progress(areas, '连接段数据保存中（1/2）'):
        junction = Junction(ConnectorArea)
        junctions.append(junction)
        for connector in ConnectorArea.allConnector():
            # 为所有的车道连接创建独立的 road，关联至 junction
            for laneConnector in connector.laneConnectors():
                connectors.append(Connector(laneConnector, junction))

    # 遍历路段
    roads = []
    links = netiface.links()
    for link in pgd.progress(links, '路段数据保存中（2/2）'):
        roads.append(Road(link))

    # 路网绘制成功后，写入xodr文件
    doc = Doc()
    # 如果有投影中心的信息
    if proj_string:
        doc.init_doc(proj_string=proj_string)
    else:
        doc.init_doc()
    doc.add_junction(junctions)
    doc.add_road(roads + connectors)

    uglyxml = doc.doc.toxml()
    xml = minidom.parseString(uglyxml)
    xml_pretty_str = xml.toprettyxml()
    
    return xml_pretty_str


from lxml import etree

from .utils import config


def opendrive2tess(netiface, params):
    # params:
    # - file_path
    # - step_length
    # - lane_types
    
    file_path = params["file_path"]
    step_length = params["step_length"]
    lane_types = params["lane_types"]
    
    config.sceneScale = netiface.sceneScale()

    from .opendrive2lanelet.opendriveparser.parser import parse_opendrive
    from .utils.network_utils import Network
    
    with open(file_path, "r", encoding='utf-8') as file_in:
        root_node = etree.parse(file_in).getroot()
        opendrive = parse_opendrive(root_node)
    
    network = Network(opendrive)
    network.convert_network(step_length)
    error_junction = network.create_network(lane_types, netiface)
    if error_junction:
        print("error_junction:", error_junction)




import os
import sys
import shutil
from pathlib import Path
from PySide2.QtWidgets import QApplication

from pytessng.DLLs.Tessng import TessngFactory
from pytessng.Tessng.MyPlugin import MyPlugin


class TessngObject:
    def __init__(self, extension=False):
        # 加载之前
        self.before_load()
        # 初始化界面
        self.app = QApplication()
        self.config = {
            '__workspace': self.workspace_path, # 工作空间
            # '__netfilepath': "",
            '__simuafterload': False, # 加载路网后是否自动启动仿真
            '__custsimubysteps': False, # 是否自定义仿真调用频率
            '__allowspopup': False, # 禁止弹窗
            '__cacheid': True, # 快速创建路段
        }
        self.plugin = MyPlugin(extension)
        self.factory = TessngFactory()
        self.tessng = self.factory.build(self.plugin, self.config)
        # 启动
        self.run()

    def run(self, ):
        if self.tessng is not None:
            sys.exit(self.app.exec_())
        else:
            sys.exit()

    def before_load(self):
        # 工作空间是本进程所在的路径
        self.workspace_path = os.path.join(os.getcwd(), "WorkSpace")
        # 创建文件夹
        os.makedirs(self.workspace_path, exist_ok=True)

        # 试用版key的位置
        cert_file_path = os.path.abspath(Path(__file__).resolve().parent / "Files" / "Cert" / "JidaTraffic_key")
        # 复制key
        _cert_folder_path = os.path.join(self.workspace_path, "Cert")
        _cert_file_path = os.path.join(_cert_folder_path, "_cert")
        if not os.path.exists(_cert_file_path):
            new_cert_file_path = os.path.join(self.workspace_path, "Cert", "可使用本试用版密钥激活TESSNG")
            os.makedirs(_cert_folder_path, exist_ok=True)
            shutil.copy(cert_file_path, new_cert_file_path)

        # 导入样例的位置
        examples_file_path = os.path.abspath(Path(__file__).resolve().parent / "Files" / "Examples")
        new_examples_file_path = os.path.join(os.getcwd(), "WorkSpace", "Examples")
        try:
            shutil.copytree(examples_file_path, new_examples_file_path)
        except:
            pass



sceneScale = 1


# tess 中车道类型定义 机动车道/机非共享/非机动车道/公交专用道
LANE_TYPE_MAPPING = {
    'driving': '机动车道',
    'biking': '非机动车道',
    'sidewalk': '人行道',  # 行人道实际无意义
    'stop': '应急车道',

    # 'onRamp': '机动车道',
    # 'offRamp': '机动车道',
    # 'entry': '机动车道',
    # 'exit': '机动车道',
    # 'connectingRamp': '机动车道',
    # 'shoulder': '应急车道',
    # 'border': '',
    # 'none': '',
    # 'redtricted': '',
    # 'parking': '停车带',
    # 'median': '',
    # 'curb': '',
}

# 当前后几个点的向量夹角小于 default_angle 且点距小于 max_length(除非夹角为0 ) 时，抹除过渡点
default_angle = 1
max_length = 50


# 连续次数后可视为正常车道，或者连续次数后可视为连接段,最小值为2
point_require = 2
POINT_REQUIRE = max(2, point_require)



# 需要被处理的车道类型及处理参数
WIDTH_LIMIT = {
    '机动车道': {
        'split': 2,  # 作为正常的最窄距离
        'join': 1.5,  # 被忽略时的最宽距离
    },
    '非机动车道': {
        'split': 2,
        'join': 0.5,
    },
}




import collections
from collections import defaultdict
from typing import List, Dict
from numpy import linspace, sqrt
from commonroad.scenario.scenario import Scenario

from ..opendrive2lanelet.network import Network
from ..opendrive2lanelet.opendriveparser.elements.opendrive import OpenDrive
from ..opendrive2lanelet.opendriveparser.elements.roadLanes import LaneSection
from ..opendrive2lanelet.opendriveparser.elements.roadElevationProfile import ElevationRecord
try:
    from .....Tessng.MyMenu import ProgressDialog as pgd
except:
    class pgd: progress = lambda values, text="": values


def convert_opendrive(opendrive: OpenDrive, filter_types: List, roads_info: Dict, context: Dict = None) -> Scenario:
    """
        将路网模型转换为 Scenario 对象
    Args:
        opendrive: 
        filter_types: 需提取的 opendrive 车道类型列表
        roads_info: 已被转换的路段详情字典
        context: 上下文对象

    Returns:

    """
    road_network = Network()
    road_network.load_opendrive(opendrive)
    return road_network.export_commonroad_scenario(filter_types=filter_types, roads_info=roads_info, context=context)


def calc_elevation(pos: float, elevations: List[ElevationRecord]):
    """
     计算路网任意一点的高程
    Args:
        pos: 距离路段起点位置
        elevations: 已排序的高程元素列表

    Returns:
     高程
    """
    if not elevations:
        return 0  # 无高程元素，默认高度为0
    # 获取相应的 elevation
    for e in elevations:
        if pos >= e.start_pos:
            elevation = e
    a, b, c, d = elevation.polynomial_coefficients
    ds = pos - elevation.start_pos  # 每当新的元素出现，`ds`则清零
    high = a + b * ds + c * ds ** 2 + d * ds ** 3
    return high


def calc_width(l1: List, l2: List):
    """
        通过两列表相应位置的点对比计算宽度序列
    """
    width_list = []
    for index in range(len(l1)):
        width = sqrt((l1[index][0] - l2[index][0]) ** 2 + (l1[index][1] - l2[index][1]) ** 2)
        width_list.append(width)
    return width_list


def convert_roads_info(opendrive: OpenDrive, step_length: int,
                       filter_types: List[str]) -> Dict:  # step_length需要对第三方包进行修改
    """
        对路段级别信息进行二次处理，获取有效信息，将其保存为字典
    Args:
        opendrive: 
        step_length: 
        filter_types: 

    Returns:

    """
    roads_info = {}
    for road in pgd.progress(opendrive.roads, '数据解析中（1/13）'):
        road_length = road.length
        planView = road.planView  # 每条道路有且仅有一条参考线，参考线通常在道路中心，但也可能有侧向偏移。
        road_points = {}

        # 计算高程信息
        elevations = [elevation[0] for elevation in road.elevationProfile.elevations]
        # 为了适配tess，将road按照section切分为多个link
        for section in road.lanes.lane_sections:
            section_id = section.idx
            section_length = section.length
            section_sPos = section.sPos
            section_ePos = section_sPos + section_length

            steps = int(section_length // step_length + 2)  # steps >= 2
            lengths = list(linspace(section_sPos, section_ePos, steps))
            points = []
            # offsets = []
            # right_offsets = []
            # lengths = list(road_section_distance[road.id][section_id].values())[0] # 尽量拟合路段与车道
            # 计算每一点的坐标和角度
            for length in lengths:
                # 根据点位 计算左右来向相应点的坐标/角度/高程
                position, angle = planView.calc_geometry(length)
                elevation_result = calc_elevation(length, elevations)

                points.append(
                    {
                        "position": list(position) + [elevation_result],
                        'angle': angle,
                        "offset": length,  # 记录在本section内此点的移动位置
                    }
                )

            # 左右方向参考线点计算不一样
            road_points[section_id] = {
                "right_points": points,
                "left_points": points[::-1],
                'sPos': section_sPos,
                'ePos': section_ePos,
                'length': section_length,
                'steps': steps,
                'lengths': lengths,
                "left_offsets": [road_length - offset for offset in lengths],
                "right_offsets": lengths,
                "elevations": [],
            }

        # # 计算每一段section 的高程信息
        # # 获取高程分段列表
        # elevations = [elevation[0] for elevation in road.elevationProfile.elevations]
        # for section_id, section_info in road_points.items():
        #     sPos = section_info["sPos"]
        #     ePos = section_info["ePos"]
        #     section_info["start_high"] = calc_elevation(sPos, elevations)
        #     section_info["end_high"] = calc_elevation(ePos, elevations)
        #     for index, length in enumerate(section_info["lengths"]):
        #         section_info["elevations"].append(calc_elevation(length, elevations))

        sections_mapping = convert_section_info(road.lanes.lane_sections, filter_types)
        # elevations [(e1),(e2),(e3)]  start_pos, road.elevationProfile.elevations[0][0].polynomial_coefficients
        roads_info[road.id] = {
            "name": road.name,
            "junction_id": road.junction and road.junction.id,  # -1 为非junction，此道路是在交叉口内部
            'road_points': road_points,  # 每个section 分开，方便微观处理,每条lane点位详情
            'length': road_length,
            'lane_sections': sections_mapping,  # lane 概况
        }

    return roads_info


def convert_lanes_info(opendrive: OpenDrive, scenario: Scenario, roads_info: Dict) -> Dict:
    """
        对车道级别信息进行二次处理，获取有效信息，将其保存为字典
    Args:
        opendrive: 
        scenario: 
        roads_info: 路段信息字典

    Returns:
        转换后的车道信息
    """
    # 获取 link与交叉口关系
    scenario_mapping = {
        "roads": {},
        "sections": {},
        "lanes": {},
    }
    for road in pgd.progress(opendrive.roads, '数据解析中（4/13）'):
        scenario_mapping["roads"][road.id] = road
        for sectionidx, section in enumerate(road.lanes.lane_sections):
            scenario_mapping["sections"][f"{road.id},{sectionidx}"] = section
            for lane in section.allLanes:
                scenario_mapping["lanes"][f"{road.id},{sectionidx},{lane.id}"] = lane
                if lane.id == 0:  # 中心车道信息保存在road info中
                    roads_info[road.id]['lane_sections'][sectionidx]['center_lane'] = {
                        "lane_id": lane.id,
                        "road_marks": lane.road_marks,
                        "widths": lane.widths,
                    }

    # 获取道路与路段关系
    lanes_info = defaultdict(dict)
    # 中心车道未转换，用参考线代替
    for lane in pgd.progress(scenario.lanelet_network.lanelets, '数据解析中（5/13）'):
        # center_lane
        # 获取所在路段
        lane_name = lane.lanelet_id
        ids = lane_name.split('.')
        road_id = int(ids[0])
        section_id = int(ids[1])
        lane_id = int(ids[2])
        road_marks = scenario_mapping["lanes"][f"{road_id},{section_id},{lane_id}"].road_marks

        # 计算车道宽度
        center_vertices, left_vertices, right_vertices = lane.center_vertices.tolist(), lane.left_vertices.tolist(), lane.right_vertices.tolist()
        widths = calc_width(left_vertices, right_vertices)

        # 添加高程
        if lane_id > 0:
            elevtions = [i["position"][2] for i in roads_info[road_id]["road_points"][section_id]['left_points']]
        else:
            elevtions = [i["position"][2] for i in roads_info[road_id]["road_points"][section_id]['right_points']]
        center_vertices = [list(center_vertices[index]) + [elevtions[index]] for index in range(len(elevtions))]
        left_vertices = [list(left_vertices[index]) + [elevtions[index]] for index in range(len(elevtions))]
        right_vertices = [list(right_vertices[index]) + [elevtions[index]] for index in range(len(elevtions))]

        # lane.lanelet_id 自定义的车道编号,取消转换后，指的就是原始编号
        lanes_info[lane.lanelet_id] = {
            "road_id": road_id,
            "section_id": section_id,
            "lane_id": lane_id,
            "left": {
                "lane_id": lane.adj_left,
                "same_direction": lane.adj_left_same_direction,
            },
            "right": {
                "lane_id": lane.adj_right,
                "same_direction": lane.adj_right_same_direction,
            },
            "predecessor_ids": lane.predecessor,
            "successor_ids": lane.successor,
            "type": lane.type,
            "name": lane_name,  # road_id+lane_section+lane_id+-1
            "center_vertices": center_vertices,
            "left_vertices": left_vertices,
            "right_vertices": right_vertices,
            "widths": widths,
            "road_marks": road_marks,
            'traffic_lights': list(lane.traffic_lights),
            'traffic_signs': list(lane.traffic_signs),
            'distance': list(lane.distance),
        }

    # 车道ID，中心车道为0， 正t方向升序，负t方向降序(基本可理解为沿参考线从左向右下降)
    return lanes_info


def convert_section_info(sections: LaneSection, filter_types: List[str]):
    """
    按section级别提取车道信息
    Args:
        sections: 所有的 section 元素
        filter_types: 需要被提取的opendrive车道类型

    Returns:

    """
    sections_mapping = collections.defaultdict(lambda: {
        'right': [],
        'center': [],
        'left': [],
        'all': [],
        'infos': {}
    }
                                               )
    for section in sections:
        section_id = sections.index(section)
        for lane in section.allLanes:
            if filter_types is not None and lane.type not in filter_types:
                continue
            if lane.id == 0:
                direction = 'center'
            elif lane.id >= 0:
                direction = 'left'
            else:
                direction = 'right'
            sections_mapping[section_id][direction].append(lane.id)
        sections_mapping[section_id]['all'] = sections_mapping[section_id]['right'] + sections_mapping[section_id][
            'left'] + sections_mapping[section_id]['center']
    return sections_mapping


# 提取车道限制信息
def lane_restrictions(lane):
    speed_info = []
    access_info = []
    for speed in lane.getElementsByTagName('speed'):
        speed_info.append(
            {
                "sOffset ": speed.getAttribute("sOffset"),
                "max": float(speed.getAttribute("max")),
                "unit": speed.getAttribute("unit"),
            }
        )
    for access in lane.getElementsByTagName('access'):
        access_info.append(
            {
                "sOffset ": access.getAttribute("sOffset"),
                "rule ": access.getAttribute("rule"),
                "restriction ": access.getAttribute("restriction"),
            }
        )
    return {
        "speeds": speed_info,
        'accesss': access_info,
    }



import os
import json
from .utils import config


def tess2shape(netiface, params, mode):
    # params:
    # - file_path
    # - proj_string

    # 文件保存路径
    file_path = params["file_path"]
    
    # 场景比例尺
    config.sceneScale = netiface.sceneScale()
    
    # 导入计算模块
    from .utils.get_data import get_info

    # 获取数据
    lane_gdf, connector_gdf = get_info(netiface, params["proj_string"])
    
    # 写入数据
    if mode == "shape":
        # 创建文件夹
        os.makedirs(file_path, exist_ok=True)
        # 写入数据
        lane_gdf.to_file(os.path.join(file_path, "lane.shp"))
        if connector_gdf is not None:
            connector_gdf.to_file(os.path.join(file_path, "laneConnector.shp"))
    elif mode == "geojson":
        lane_geojson = json.loads(lane_gdf.to_json())
        # 将车道的type改为lane
        for i in range(len(lane_geojson["features"])):
            lane_geojson["features"][i]["type"] = "lane"
        # 将车道连接的type改为laneConnector
        if connector_gdf is not None:
            connector_geojson = json.loads(connector_gdf.to_json())
            for i in range(len(connector_geojson["features"])):
                connector_geojson["features"][i]["type"] = "laneConnector"
            lane_geojson["features"].extend(connector_geojson["features"])
        # 写入数据
        with open(file_path, "w", encoding="utf-8") as json_file:
            json.dump(lane_geojson, json_file, indent = 4, ensure_ascii=False)



# -*- coding: utf-8 -*-

"""Module to enhance Lanelet class so it can be used
for conversion from the opendrive format."""

from typing import Tuple

import numpy as np
from commonroad.scenario.lanelet import Lanelet
from commonroad.scenario.lanelet import LaneletType

__author__ = "Benjamin Orthen, Sebastian Maierhofer"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"

optimal_join_split_factor = 20


class ConversionLanelet(Lanelet):
    """Change some properties of the Lanelet class so that it can be used
    to conversions to Lanelet. This means especially that lanelet_ids
    can be other types than a natural number and that these ids can be changed
    more than once.
    Also adjacent neighbors and pre- and successor can be changed more than onc.

    Args:

    Returns:

    """

    def __init__(
            self,
            parametric_lane_group,
            left_vertices: np.ndarray,
            center_vertices: np.ndarray,
            right_vertices: np.ndarray,
            lanelet_id,
            predecessor=None,
            successor=None,
            adjacent_left=None,
            adjacent_left_same_direction=None,
            adjacent_right=None,
            adjacent_right_same_direction=None,
            line_marking_left_vertices=None,
            line_marking_right_vertices=None,
            stop_line=None,
            lanelet_type=None,
            user_one_way=None,
            user_bidirectional=None,
            traffic_signs=None,
            traffic_lights=None,
    ):
        if lanelet_type is None:
            lanelet_type = {LaneletType.UNKNOWN}
        super().__init__(
            left_vertices=left_vertices,
            center_vertices=center_vertices,
            right_vertices=right_vertices,
            lanelet_id=1,
            predecessor=predecessor,
            successor=successor,
            adjacent_left=adjacent_left,
            adjacent_left_same_direction=adjacent_left_same_direction,
            adjacent_right=adjacent_right,
            adjacent_right_same_direction=adjacent_right_same_direction,
            line_marking_left_vertices=line_marking_left_vertices,
            line_marking_right_vertices=line_marking_right_vertices,
            stop_line=stop_line,
            lanelet_type=lanelet_type,
            user_one_way=user_one_way,
            user_bidirectional=user_bidirectional,
            traffic_signs=traffic_signs,
            traffic_lights=traffic_lights,
        )

        self.parametric_lane_group = parametric_lane_group
        self.lanelet_id = lanelet_id

    def __eq__(self, lanelet: "ConversionLanelet") -> bool:
        """Lanelets are equal if their id_ is equal.

        Args:
           lanelet: Lanelet to be compared to equality.
        Returns:
           True if id_ is equal.
        """
        return self.lanelet_id == lanelet.lanelet_id

    @property
    def lanelet_id(self) -> int:
        """Get or set id of this lanelet."""
        return self._lanelet_id

    @lanelet_id.setter
    def lanelet_id(self, l_id: int):
        # pylint: disable=W0201
        self._lanelet_id = l_id

    @property
    def left_vertices(self) -> np.ndarray:
        """Get or set right vertices of this lanelet."""
        return self._left_vertices

    @left_vertices.setter
    def left_vertices(self, polyline: np.ndarray):
        # pylint: disable=W0201
        self._left_vertices = polyline

    @property
    def right_vertices(self) -> np.ndarray:
        """Get or set right vertices of this lanelet."""
        return self._right_vertices

    @right_vertices.setter
    def right_vertices(self, polyline: np.ndarray):
        # pylint: disable=W0201
        self._right_vertices = polyline

    @property
    def center_vertices(self) -> np.ndarray:
        """Get or set center vertices of this lanelet."""
        return self._center_vertices

    @center_vertices.setter
    def center_vertices(self, polyline: np.ndarray):
        # pylint: disable=W0201
        self._center_vertices = polyline

    @property
    def predecessor(self) -> list:
        """Set or get the predecessor."""
        return self._predecessor

    @predecessor.setter
    def predecessor(self, predecessor: list):
        # pylint: disable=W0201
        self._predecessor = predecessor

    @property
    def successor(self) -> list:
        return self._successor

    @successor.setter
    def successor(self, successor: list):
        # pylint: disable=W0201
        self._successor = successor

    @property
    def adj_left(self) -> int:
        """Set or get adjacent left lanelet."""
        return self._adj_left

    @adj_left.setter
    def adj_left(self, l_id: int):
        # pylint: disable=W0201
        self._adj_left = l_id

    @property
    def adj_left_same_direction(self) -> bool:
        """Set or get if adjacent left lanelet has the same direction
        as this lanelet."""
        return self._adj_left_same_direction

    @adj_left_same_direction.setter
    def adj_left_same_direction(self, same: bool):
        # pylint: disable=W0201
        self._adj_left_same_direction = same

    @property
    def adj_right(self) -> int:
        """Set or get adjacent right lanelet."""
        return self._adj_right

    @adj_right.setter
    def adj_right(self, l_id: int):
        self._adj_right = l_id

    @property
    def adj_right_same_direction(self) -> bool:
        """Set or get if adjacent right lanelet has the same direction
        as this lanelet."""
        return self._adj_right_same_direction

    @adj_right_same_direction.setter
    def adj_right_same_direction(self, same: bool):
        # pylint: disable=W0201
        self._adj_right_same_direction = same

    def concatenate(
            self, lanelet_conc: "ConversionLanelet", extend_plane_group: bool = True
    ) -> "ConversionLanelet":
        """Concatenate this lanelet with lanelet_conc and assign the
        new lanelet_id to the resulting lanelet.

        Args:
          lanelet_conc: Lanelet which will be included.
          extend_plane_group: Whether to extend the parametric_lane_group of this lanelet
        with the parametric lanes of the lanelet_conc.parametric_lane_group.
          lanelet_conc: "ConversionLanelet":
          # lanelet_id: str:  (Default value = -1)
          extend_plane_group: bool:  (Default value = True)

        Returns:

        """

        # check connectedness
        if np.isclose(self.left_vertices[-1], lanelet_conc.left_vertices[0]).all():
            idx = 1
        else:
            idx = 0

        self.left_vertices = np.vstack(
            (self.left_vertices, lanelet_conc.left_vertices[idx:])
        )
        self.center_vertices = np.vstack(
            (self.center_vertices, lanelet_conc.center_vertices[idx:])
        )
        self.right_vertices = np.vstack(
            (self.right_vertices, lanelet_conc.right_vertices[idx:])
        )
        if extend_plane_group:
            self.parametric_lane_group.extend(
                lanelet_conc.parametric_lane_group.parametric_lanes
            )
        self.successor = lanelet_conc.successor.copy()

    def calc_width_at_end(self) -> float:
        """Calc width of lanelet at its end.

        Returns:
          Width at end of lanelet.

        """
        return self.calc_width(self.length)

    def calc_width_at_start(self) -> float:
        """Calc width of lanelet at its start.

        Returns:
          Width at start of lanelet.

        """
        return self.calc_width(0)

    def calc_width(self, s_pos: float) -> float:
        """Calc width at position s_pos.

        Args:
          s_pos: Position in curve parameter ds.
        Returns:
          Width at postiion s_pos.
        """
        inner_pos = self.calc_border("inner", s_pos)[0]
        outer_pos = self.calc_border("outer", s_pos)[0]
        return np.linalg.norm(inner_pos - outer_pos)

    @property
    def length(self) -> float:
        """Get length of lanelet by calculating length of ParametricLaneGroup.

        Returns:
          Length of lanelet.
        """
        return self.parametric_lane_group.length

    def has_zero_width_everywhere(self) -> bool:
        """Checks if width is zero at every point of its ParametricLaneGroup.

        Returns:
          True if every ParametricLane has width_coefficients equal to only zero.
        """
        return self.parametric_lane_group.has_zero_width_everywhere()

    def first_zero_width_change_position(
            self, reverse: bool = False, reference_width: float = 0.0
    ) -> float:
        """Get the earliest point of the lanelet where the width change is zero.

        Args:
          reverse: True if checking starts from the end of the lanelet.
          reference_width: Width for which width at zero width change position has
            to be greater as.
        Returns:
          Position of lanelet (in curve parameter ds) where width change is zero.
        """
        return self.parametric_lane_group.first_zero_width_change_position(
            reverse, reference_width
        )

    def maximum_width(self) -> float:
        return self.parametric_lane_group.maximum_width()

    def optimal_join_split_values(
            self, is_split: bool, split_and_join: bool, reference_width: float
    ):
        """Calculate an optimal value, where the lanelet split or join starts
          or ends, respectively.

        Args:
          is_split: True if lanelet splits from another lanelet, otherwise
            False if it is a join.
          split_and_join: True if lanelet has a split at the start and join at the end.
          reference_width: Width for which width at zero width change position has
            to be greater as.
        """

        merge_pos, merge_width = self.first_zero_width_change_position(
            reverse=(not is_split), reference_width=reference_width
        )
        if merge_pos is None:
            merge_pos = self.length if is_split else 0
            # merge_width = self.calc_width(merge_pos)

        if is_split:
            # if both a split at the start and merge at the end
            if split_and_join and merge_pos > 0.5 * self.length:
                merge_pos = 0.45 * self.length

        else:
            if split_and_join and merge_pos < 0.5 * self.length:
                merge_pos = 0.55 * self.length

        merge_width = self.calc_width(merge_pos)
        return merge_pos, merge_width

    def move_border(
            self,
            mirror_border: str,
            mirror_interval: Tuple[float, float],
            distance: Tuple[float, float],
            adjacent_lanelet: "ConversionLanelet",
    ):
        """Move vertices of one border by mirroring other border with
        a specified distance.

        Args:
          mirror_border: Which border to mirror, either 'left' or 'right'.
          interval: Tuple of two values, specifying start and end of mirroring.
          distance: Specifying distance at start and at end of mirroring.
        """
        if mirror_border == "left":
            distance[:] = [-1 * x for x in distance]

        lanelet = self.parametric_lane_group.to_lanelet_with_mirroring(
            mirror_border=mirror_border,
            distance=distance,
            mirror_interval=mirror_interval,
            adjacent_lanelet=adjacent_lanelet,
        )

        self.left_vertices = lanelet.left_vertices
        self.center_vertices = lanelet.center_vertices
        self.right_vertices = lanelet.right_vertices

    def calc_border(self, border: str, s_pos: float, width_offset: float = 0.0):
        """Calc border position according to parametric_lane_group.

        Note: This does not consider borders which have been moved
         due to joining / splitting.

        Args:
          border: Which border to calculate (inner or outer).
          s_pos: Position of parameter ds where to calc the
            Cartesian coordinates
          width_offset: Offset to add to calculated width in reference
           to the reference border. (Default value = 0.0)

        Returns:
          Cartesian coordinates of point on inner border
            and tangential direction, too.
        """
        return self.parametric_lane_group.calc_border(border, s_pos, width_offset)



import json
from .utils import config


def tess2json(netiface, params):
    # params: 参数
    # - file_path
    # - proj_string
    
    # 文件保存路径
    file_path = params["file_path"]
    
    # 场景比例尺
    config.sceneScale = netiface.sceneScale()
    
    # 导入计算模块
    from .utils.get_data import get_data
    
    # 获取数据
    json_data = get_data(netiface, params["proj_string"])
    
    # 写入数据
    with open(file_path, "w", encoding="utf-8") as json_file:
        json.dump(json_data, json_file, indent=4, ensure_ascii=False)



from kafka import KafkaProducer


# kafka 消息生产者
class KafkaMessageProducer:
    def __init__(self, bootstrap_servers:str, topic:str):
        self.producer = KafkaProducer(
            bootstrap_servers = bootstrap_servers,
            api_version = (0, 10),
            max_request_size = 20 * 1024 * 1024,
            acks = 0,  # 不等待确认
            max_in_flight_requests_per_connection=100,  # 控制最大同时未确认的请求
        )
        self.topic = topic

    # 发数据
    def send_message(self, message:str):
        try:
            self.producer.send(self.topic, message.encode('utf-8'))
            print(f"Kafka 消息发送成功.")
            return True
        except Exception as e:
            print(f"Kafka 发送消息时发生错误：{e}")
            return False

    # 关闭
    def close(self):
        self.producer.close()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Script to call from commandline to convert OSM to a CommonRoad file."""

import os
import sys
import argparse
from lxml import etree

from commonroad.common.file_writer import CommonRoadFileWriter, OverwriteExistingFile
from commonroad.planning.planning_problem import PlanningProblemSet
from commonroad.scenario.scenario import Tag
from commonroad.common.file_reader import CommonRoadFileReader

from opendrive2tessng.opendrive2lanelet.osm.osm2lanelet import OSM2LConverter
from opendrive2tessng.opendrive2lanelet.osm.parser import OSMParser
from opendrive2tessng.opendrive2lanelet.osm.lanelet2osm import L2OSMConverter

__author__ = "Benjamin Orthen"
__copyright__ = "TUM Cyber-Physical Systems Group"
__credits__ = ["Priority Program SPP 1835 Cooperative Interacting Automobiles"]
__version__ = "1.2.1"
__maintainer__ = "Sebastian Maierhofer"
__email__ = "commonroad-i06@in.tum.de"
__status__ = "Released"


def parse_arguments():
    """Parse command line arguemts.

    Returns:
      args object
    """
    parser = argparse.ArgumentParser(
        description="Parse an OSM file containing lanelets and convert it to a CommonRoad scenario file or vice versa."
    )
    parser.add_argument("input_file", help="file to be converted")
    parser.add_argument(
        "-f",
        "--force-overwrite",
        action="store_true",
        help="overwrite existing file if it has same name as converted file",
    )
    parser.add_argument("-o", "--output-name", help="specify name of outputed file")
    parser.add_argument("--proj", help="proj-string to specify conversion")
    parser.add_argument(
        "-a",
        "--adjacencies",
        action="store_true",
        help="detect left and right adjacencies of lanelets if they do not share a common way",
    )
    parser.add_argument(
        "-r",
        "--reverse",
        action="store_true",
        help="convert in the reverse direction, i.e. from CommonRoad to OSM lanelets.",
    )
    parser.add_argument(
        "--left-driving",
        action="store_true",
        help="set to true if map describes a left driving system (e.g. in Great Britain)",
    )
    args = parser.parse_args()
    return args


def main():
    """Run script to convert file."""
    args = parse_arguments()
    partition = args.input_file.rpartition(".")

    file_ending = "osm" if args.reverse else "xml"

    if args.output_name:
        output_name = args.output_name
    else:
        output_name = f"{partition[0]}.{file_ending}"  # only name of file

    if os.path.isfile(output_name) and not args.force_overwrite:
        print(
            "Not converting because file exists and option 'force-overwrite' not active",
            file=sys.stderr,
        )
        sys.exit(-1)

    if args.reverse:
        commonroad_to_osm(args, output_name)
    else:
        osm_to_commonroad(args, output_name)


def osm_to_commonroad(args, output_name: str):
    """Convert from OSM to CommonRoad.
    Args:
      args: Object containing parsed arguments.
      output_name: Name of file where result should be written to.
    """
    with open(f"{args.input_file}", "r") as file_in:
        parser = OSMParser(etree.parse(file_in).getroot())
        osm = parser.parse()

    osm2l = OSM2LConverter(proj_string=args.proj)
    scenario = osm2l(
        osm, detect_adjacencies=args.adjacencies, left_driving_system=args.left_driving
    )
    if scenario:
        writer = CommonRoadFileWriter(
            scenario=scenario,
            planning_problem_set=PlanningProblemSet(),
            author="",
            affiliation="",
            source="OpenDRIVE 2 Lanelet Converter",
            tags={Tag.URBAN, Tag.HIGHWAY},
        )
        writer.write_to_file(output_name, OverwriteExistingFile.ALWAYS)
    else:
        print("Could not convert from OSM to CommonRoad format!")


def commonroad_to_osm(args, output_name: str):
    """Convert from CommonRoad to OSM.

    Args:
      args: Object containing parsed arguments.
      output_name: Name of file where result should be written to.
    """
    try:
        commonroad_reader = CommonRoadFileReader(args.input_file)
        scenario, _ = commonroad_reader.open()

    except etree.XMLSyntaxError as xml_error:
        print(f"SyntaxERror: {xml_error}")
        print(
            "There was an error during the loading of the selected CommonRoad file.\n"
        )
    l2osm = L2OSMConverter(args.proj)
    osm = l2osm(scenario)
    with open(f"{output_name}", "wb") as file_out:
        file_out.write(
            etree.tostring(
                osm, xml_declaration=True, encoding="UTF-8", pretty_print=True
            )
        )



# -*- coding: utf-8 -*-


class RoadType:
    """ """

    allowedTypes = [
        "unknown",
        "rural",
        "motorway",
        "town",
        "lowSpeed",
        "pedestrian",
        "bicycle",
    ]

    def __init__(self, s_pos=None, use_type=None, speed=None):
        self.start_pos = s_pos
        self.use_type = use_type
        self.speed = speed

    @property
    def start_pos(self):
        """ """
        return self._sPos

    @start_pos.setter
    def start_pos(self, value):
        """

        Args:
          value:

        Returns:

        """
        # pylint: disable=W0201
        self._sPos = float(value)

    @property
    def use_type(self):
        """ """
        return self._use_type

    @use_type.setter
    def use_type(self, value):
        """

        Args:
          value:

        Returns:

        """
        if value not in self.allowedTypes:
            raise AttributeError("Type not allowed.")
        # pylint: disable=W0201
        self._use_type = value

    @property
    def speed(self):
        """ """
        return self._speed

    @speed.setter
    def speed(self, value):
        """

        Args:
          value:

        Returns:

        """
        if not isinstance(value, Speed) and value is not None:
            raise TypeError("Value {} must be instance of Speed.".format(value))
        # pylint: disable=W0201
        self._speed = value


class Speed:
    """ """

    def __init__(self, max_speed=None, unit=None):
        self._max = max_speed
        self._unit = unit

    @property
    def max(self):
        """ """
        return self._max

    @max.setter
    def max(self, value):
        """

        Args:
          value:

        Returns:

        """
        self._max = str(value)

    @property
    def unit(self):
        """ """
        return self._unit

    @unit.setter
    def unit(self, value):
        """

        Args:
          value:

        Returns:

        """
        # TODO validate input
        self._unit = str(value)



import json
import sqlite3
from typing import Dict, List
import collections
import numpy as np
from PySide2.QtGui import QVector3D
from .config import *
from .functions2 import get_new_point_indexs

###############################################################################

def p2m(x):
    return x * sceneScale

def m2p(x):
    return x / sceneScale

###############################################################################

def qtpoint2point(qtpoints):
    points = []
    for qtpoint in qtpoints:
        points.append(
            [m2p(qtpoint.x()), - m2p(qtpoint.y()), m2p(qtpoint.z())] if isinstance(qtpoint, QVector3D) else qtpoint
        )
    return points


def point2qtpoint(points):
    qtpoints = []
    for point in points:
        qtpoints.append(
            QVector3D(p2m(point[0]), - p2m(point[1]), p2m(point[2])) if not isinstance(point, QVector3D) else point
        )
    return qtpoints


###############################################################################

class Road:
    def __init__(self):
        self.link = None
        self.last_links = []
        self.next_links = []
        self.connectors = []  # 暂时没用到


class AdjustNetwork:
    def __init__(self, netiface):
        self.netiface = netiface
        self.roads = self.calc_connector()

        self.connector_area_mapping = self.calc_connector_area()

        # 在调整的过程中，旧的link在不断删除，同时新的link被创建，所以需要建立映射关系表
        # connector也有可能在消失，所以不再被记录，通过上下游link获取connector
        self.old_new_link_mapping = {}
        # 初始化映射表,link在调整前后可能存在一对多/多对一关系，所以用 dict&list 记录
        for link in netiface.links():
            self.old_new_link_mapping[link.id()] = [link.id()]

    # 记录全域的连接段面域
    def calc_connector_area(self):
        connector_area_mapping = collections.defaultdict(list)
        for ConnectorArea in self.netiface.allConnectorArea():
            for connector in ConnectorArea.allConnector():
                connector_area_mapping[ConnectorArea.id()].append(connector.id())
        return connector_area_mapping


    # 记录全局的连接关系
    def calc_connector(self):
        roads = {}
        for connector in self.netiface.connectors():
            # 上下游不可能为空
            last_link = connector.fromLink()
            next_link = connector.toLink()

            last_road = roads.get(last_link.id(), Road())
            next_road = roads.get(next_link.id(), Road())

            last_road.link = last_link
            last_road.next_links.append(next_link.id())
            last_road.connectors.append(connector)

            next_road.link = next_link
            next_road.last_links.append(last_link.id())
            next_road.connectors.append(connector)

            roads[next_link.id()] = next_road
            roads[last_link.id()] = last_road

        # 记录无连接段的路段
        for link in self.netiface.links():
            if link.id() not in roads:
                road = Road()
                road.link = link
                roads[link.id()] = road
        return roads

    # 计算路段上所有的路段点，车道点
    # 根据已知的信息计算新的断点序列列表
    @staticmethod
    def calc_points(points, indexs_list, ratios):
        last_index = 0
        points = qtpoint2point(points)

        new_points_list = []  # [[] for _ in indexs_list]

        first_point = None
        # 先把初识点全部分配下去
        for _, indexs in enumerate(indexs_list):
            if indexs:
                last_index = indexs[-1]  # 更新计算点
            new_points = [points[index] for index in indexs]
            # 添加首尾点
            if first_point is not None:
                new_points.insert(0, first_point)

            ratio = ratios[_]
            final_point = np.array(points[last_index]) * (1 - ratio) + np.array(points[last_index + 1]) * ratio
            new_points.append(final_point)
            first_point = final_point

            new_points_list.append(new_points)
            # if indexs:
            #     last_index = indexs[-1]  # 更新计算点

        # 添加最后一个路段，第一个点为上一路段计算的终点，第二个点为上一路段所在点序列的下一个点，最后一个点为路段终点
        new_points_list.append([first_point] + points[last_index + 1:])
        new_points_list = [point2qtpoint(i) for i in new_points_list]
        return new_points_list

    def calc_split_parameter(self, link, split_link_info):
        center_points = link.centerBreakPoint3Ds()
        center_points = qtpoint2point(center_points)

        sum_length = 0
        last_x, last_y, last_z = center_points[0]
        lengths = [0] + split_link_info["lengths"]
        for point_index, point in enumerate(center_points):
            x, y, z = point
            distance = np.sqrt((x - last_x) ** 2 + (y - last_y) ** 2)
            last_x, last_y, last_z = x, y, z

            new_sum_length = sum_length + distance
            for split_index, split_length in enumerate(lengths):
                if split_index == 0:
                    continue
                if new_sum_length < lengths[split_index - 1]:
                    continue
                elif new_sum_length >= lengths[split_index - 1] and new_sum_length < lengths[split_index]:
                    # 区间命中
                    split_link_info['index'][split_index - 1].append(point_index)
                    # TODO 区间已经命中，可以 break 了
                # elif sum_length >= lengths[split_index - 1] and new_sum_length > split_length:
                #     # 新点超出范围,计算比例
                #     ratio = (split_length - sum_length) / distance
                #     split_link_info['ratio'][split_index - 1] = ratio  # 此处不需要记录 首点的比例，因为当需要用到首点时，采用上一点的断点计算
                else:  # new_sum_length >= lengths[split_index]
                    # 如果此处并没有 ratio，说明这是第一次被匹配上，需要进行ratio 计算
                    if split_link_info['ratio'][split_index - 1] is None:
                        ratio = (split_length - sum_length) / distance
                        split_link_info['ratio'][split_index - 1] = ratio  # 此处不需要记录 首点的比例，因为当需要用到首点时，采用上一点的断点计算
            sum_length = new_sum_length

        # 计算完成，可以进行分割
        if len(split_link_info['lengths']) != len(split_link_info['index']) or len(split_link_info['lengths']) != len(
                split_link_info['index']):
            raise 1  # 初步判断

    def calc_split_links_info(self, link, split_link_info):
        # 计算新的 link 信息
        indexs_list = split_link_info['index']
        ratios = split_link_info['ratio']
        link_center_points = self.calc_points(link.centerBreakPoint3Ds(), indexs_list, ratios)

        new_links_info = [
            {
                'center': link_center_points[index],
                # 'name': f"{link.name()}/{index}",
                'name': f"{link.name()}",  # 保留原路段名
                'lanes': collections.defaultdict(lambda: {
                    'center': [],
                    'left': [],
                    'right': [],
                    'type': '',
                    'attr': {},
                }),
                'old_link_id': link.id(),
            } for index in range(len(indexs_list) + 1)
        ]

        for lane in link.lanes():
            center_points = self.calc_points(lane.centerBreakPoint3Ds(), indexs_list, ratios)
            left_points = self.calc_points(lane.leftBreakPoint3Ds(), indexs_list, ratios)
            right_points = self.calc_points(lane.rightBreakPoint3Ds(), indexs_list, ratios)
            for index in range(len(indexs_list) + 1):  # 被分割后的 link 数量,比分割点数大 1
                new_links_info[index]['lanes'][lane.number()] = {
                    'center': center_points[index],
                    'left': left_points[index],
                    'right': right_points[index],
                    'type': lane.actionType(),
                    'attr': {},
                }
        return new_links_info

    def split_link(self, reader):
        split_links_info = collections.defaultdict(lambda: {'lengths': [], 'index': [], 'ratio': []})
        for row in reader:
            try:
                row = [float(i) for index, i in enumerate(row)]
            except:
                return f"输入数据错误:{row}"

            link_id = int(row[0])
            points_length = sorted(row[1:])
            link = self.netiface.findLink(link_id)
            if not link:
                return f"link: {link_id} 不存在"
            if min(points_length) <= 0 or max(points_length) >= link.length() or len(points_length) != len(
                    set(points_length)):
                return f"link: {row[0]} 长 {link.length()}, 断点长度输入不准确"

            split_links_info[link_id]['lengths'] = points_length
            split_links_info[link_id]['index'] = [[] for _ in points_length]
            split_links_info[link_id]['ratio'] = [None for _ in points_length]

        # 路段有可能发生 合并，切分，所以使用 list 存储最为合理
        old_new_link_mapping = collections.defaultdict(list)
        for link in self.netiface.links():
            old_new_link_mapping[link.id()].append(link.id())

        old_connectors = []
        # 记录原始连接信息
        for link_id in split_links_info.keys():
            # last_link, next_link 都不属于 link_group,所以在过程中可能已经被删除，只能通过id映射获取新的link
            # 可能存在多个上游路段和多个下游路段，需要逐个获取连接段，从而记录连接关系
            old_connectors += self.calc_upstream_connection(link_id)
            old_connectors += self.calc_downstream_connection(link_id)

        all_new_link_info = []
        # 计算新的路段信息
        for link_id, split_link_info in split_links_info.items():
            link = self.netiface.findLink(link_id)
            # 根据 link 实际信息 丰富 切割参数 > split_link_info
            self.calc_split_parameter(link, split_link_info)
            # 获取切割详情，含点序列等基本信息
            new_links_info = self.calc_split_links_info(link, split_link_info)
            all_new_link_info += new_links_info
            # 记录 link 基本信息后移除
            old_link_id = link.id()
            old_new_link_mapping[old_link_id].remove(old_link_id)  # 删除路段前，移除相关的映射关系
            self.netiface.removeLink(link)

        temp_link_mapping = collections.defaultdict(list)
        # 根据记录的信息, 集中创建新的路段并更新映射表
        for new_link_info in all_new_link_info:
            # 做路段分割时，原始路段有且仅有一个
            old_link_id = new_link_info['old_link_id']
            new_link_obj = self.create_new_link(new_link_info)
            # 记录进映射表
            old_new_link_mapping[old_link_id].append(new_link_obj.id())
            temp_link_mapping[old_link_id].append(new_link_obj.id())

        message = "路段分割结果: \n"
        for k, v in temp_link_mapping.items():
            message += f"{k} --> {v} \n"

        # 集中创建内部的连接段
        for old_link_id, new_link_ids in old_new_link_mapping.items():
            for index in range(len(new_link_ids) - 1):
                from_link_id = old_new_link_mapping[old_link_id][index]
                to_link_id = old_new_link_mapping[old_link_id][index + 1]
                link_obj = self.netiface.findLink(from_link_id)
                # 被分割的小路段的车道数/车道类型均完全一致，所以随便取一个路段即可
                lanes = [lane.number() + 1 for lane in link_obj.lanes()]
                self.netiface.createConnector(from_link_id, to_link_id, lanes, lanes, f"{from_link_id}-{to_link_id}")

        # 根据记录的信息 批量创建新的连接段
        self.create_all_new_connector(old_connectors, old_new_link_mapping)
        return message

    def join_link(self):
        # TODO 合并路段, 需要添加车道类型判断，后续增加点位自合并的方法
        link_groups = []
        exist_links = []
        for link_id, road in self.roads.items():
            if link_id in exist_links:
                # 已经进行过查找的 link 不需要再次遍历
                continue

            # 获取 link 相应的 上下游link 并组成有序列表
            link_group = [road.link]
            self.get_chain_by_next(road, link_group)
            self.get_chain_by_last(road, link_group)

            link_groups.append(link_group)
            exist_links += [i.id() for i in link_group]

        # 判断是否有路段进行过重复查询，如果有，说明逻辑存在漏洞
        if len(exist_links) != len(set(exist_links)):
            return "出现唯一性错误，请联系开发者"

        # 在调整的过程中，旧的link在不断删除，同时新的link被创建，所以需要建立映射关系表，connector也有可能在消失，所以不再被记录，通过上下游link获取connector
        # 初始化映射表, 为了和 分割路段保持一致，仍然采用 dict(list) 的形式，但实际上, 最后的 list 有且仅有一个值
        old_new_link_mapping = collections.defaultdict(list)
        for link in self.netiface.links():
            old_new_link_mapping[link.id()].append(link.id())

        # TODO 根据信息做路网调整, 在调整过程中，未遍历到的 link_group 不会被调整，即不会丢失对象
        # 分步做，先统计原始的连接段信息，方便后续迭代
        old_connectors = []
        for link_group in filter(lambda x: len(x) > 1, link_groups):
            # 记录原始信息，方便后续重新创建路段及连接段
            first_link = link_group[0]
            final_link = link_group[-1]

            # last_link, next_link 都不属于 link_group,所以在过程中可能已经被删除，只能通过id映射获取新的link
            # 可能存在多个上游路段和多个下游路段，需要逐个获取连接段，从而记录连接关系
            old_connectors += self.calc_upstream_connection(first_link.id())
            old_connectors += self.calc_downstream_connection(final_link.id())

        all_new_link_info = []
        all_remove_links = []
        # 记录路段合并信息
        for link_group in filter(lambda x: len(x) > 1, link_groups):
            new_link_info = {
                'center': [],
                'name': '',
                'lanes': collections.defaultdict(lambda: {
                    'center': [],
                    'left': [],
                    'right': [],
                    'type': '',
                    'attr': {},
                }),
                'old_link_ids': [i.id() for i in link_group],
            }

            # 先记录id
            for link in link_group:  # 有序的进行点位合并
                # TODO 暂时不记录中间连接段的点序列
                new_link_info['center'] += link.centerBreakPoint3Ds()
                new_link_info['name'] += link.name()
                for lane in link.lanes():
                    lane_number = lane.number()
                    new_link_info['lanes'][lane_number]['center'] += lane.centerBreakPoint3Ds()
                    new_link_info['lanes'][lane_number]['left'] += lane.leftBreakPoint3Ds()
                    new_link_info['lanes'][lane_number]['right'] += lane.rightBreakPoint3Ds()
                    new_link_info['lanes'][lane_number]['type'] = lane.actionType()

                # 记录 link 基本信息后移除
                # old_new_link_mapping[link.id()].remove(link.id())
                # self.netiface.removeLink(link)
            all_new_link_info.append(new_link_info)

        message = "路段合并结果: \n"
        # 集中创建新的路段并更新映射表
        for new_link_info in all_new_link_info:
            new_link_obj = self.create_new_link(new_link_info)

            # 新路段创建成功，进行老路段移除，并更新新的连接数据，否则不做处理
            if not new_link_obj:  # 新路段可能创建失败
                continue

            for old_link_id in new_link_info['old_link_ids']:
                old_link = self.netiface.findLink(old_link_id)
                old_new_link_mapping[old_link_id].remove(old_link_id)  # 旧路段的关联路段也移除
                self.netiface.removeLink(old_link)
            message += f"{new_link_info['old_link_ids']} --> {new_link_obj.id()} \n"
            # 更新映射表，原本多个旧路段都指向了同一个新路段
            for old_link_id in new_link_info['old_link_ids']:
                old_new_link_mapping[old_link_id].append(new_link_obj.id())

        # 创建新的连接段,
        # 如果某连接段上下游均进行了路段合并，则连接段会被重新重复创建，已被过滤
        self.create_all_new_connector(old_connectors, old_new_link_mapping)

        return message

    # 计算路段的上游连接关系
    def calc_upstream_connection(self, link_id):
        road = self.roads[link_id]
        connectors = []
        for last_link_id in road.last_links:
            connector_info = self.get_connector_info(last_link_id, link_id)
            connectors.append(connector_info)
        return connectors

    # 计算路段的下游连接关系
    def calc_downstream_connection(self, link_id):
        road = self.roads[link_id]
        connectors = []
        for next_link_id in road.next_links:
            connector_info = self.get_connector_info(link_id, next_link_id)
            connectors.append(connector_info)
        return connectors

    def get_connector_info(self, from_link_id, to_link_id):
        connector = self.netiface.findConnectorByLinkIds(from_link_id, to_link_id)
        return {
            'from_link_id': from_link_id,
            'to_link_id': to_link_id,
            'connector': [
                (i.fromLane().number(), i.toLane().number())
                for i in connector.laneConnectors()
            ],
            'lanesWithPoints3': [
                {
                    "center": i.centerBreakPoint3Ds(),
                    "left": i.leftBreakPoint3Ds(),
                    "right": i.rightBreakPoint3Ds(),
                }
                for i in connector.laneConnectors()
            ],
        }

    def get_chain_by_next(self, road, link_group: list):
        if len(road.next_links) != 1:
            # 有且仅有一个下游，才可以继续延伸
            return

        next_link_id = road.next_links[0]
        # 新增判断，即使路段只有一个下游，连接段所属面域中存在多个连接段，仍然不允许合并
        connector = self.netiface.findConnectorByLinkIds(road.link.id(), next_link_id)
        for value in self.connector_area_mapping.values():
            if connector.id() in value and len(value) >= 3:
                print(f"面域内连接段过多, 进入交叉口区域, 不再继续: {value}")
                return

        next_link = self.netiface.findLink(next_link_id)
        next_road = self.roads[next_link.id()]
        # 判断下游 link 是否有且仅有 1 个上游，且车道数/车道类型与当前link一致，若一致，加入链路并继续向下游寻找
        if len(next_road.last_links) == 1 and [lane.actionType() for lane in road.link.lanes()] == [
            lane.actionType() for lane in next_road.link.lanes()]:
            if next_link in link_group:
                # link已存在，说明构成了回路，直接结束
                return
            link_group.append(next_link)
            self.get_chain_by_next(next_road, link_group)
        return

    # 通过指定路段信息寻找匹配的上下游
    def get_chain_by_last(self, road, link_group: list):
        if len(road.last_links) != 1:
            # 有且仅有一个上游，才可以继续延伸
            return
        last_link_id = road.last_links[0]
        # 新增判断，即使路段只有一个下游，连接段所属面域中存在多个连接段，仍然不允许合并
        connector = self.netiface.findConnectorByLinkIds(last_link_id, road.link.id())
        for value in self.connector_area_mapping.values():
            if connector.id() in value and len(value) >= 3:
                print(f"面域内连接段过多, 进入交叉口区域, 不再继续: {value}")
                return

        last_link = self.netiface.findLink(last_link_id)
        last_road = self.roads[last_link.id()]
        # 判断上游 link 是否有且仅有 1 个下游，且车道数与当前link一致，若一致，加入链路并继续向上游寻找
        if len(last_road.next_links) == 1 and [lane.actionType() for lane in road.link.lanes()] == [
            lane.actionType() for lane in last_road.link.lanes()]:
            if last_link in link_group:
                # link已存在，说明构成了回路，直接结束
                return
            link_group.insert(0, last_link)
            self.get_chain_by_last(last_road, link_group)
        return

    def create_new_link(self, new_link_info):
        new_link_obj = self.netiface.createLink3DWithLanePointsAndAttrs(
            new_link_info['center'],
            [
                {
                    'center': new_link_info['lanes'][k]['center'],
                    'right': new_link_info['lanes'][k]['right'],
                    'left': new_link_info['lanes'][k]['left'],
                } for k in sorted(new_link_info['lanes'])
            ],  # 必须排序
            [new_link_info['lanes'][k]['type'] for k in sorted(new_link_info['lanes'])],
            [new_link_info['lanes'][k]['attr'] for k in sorted(new_link_info['lanes'])],
            new_link_info['name']
        )
        return new_link_obj

    # 根据记录的信息 批量创建新的连接段
    def create_all_new_connector(self, old_connectors, old_new_link_mapping):
        # 如果某连接段上下游均进行了路段合并，则连接段会被重新重复创建，此处进行过滤
        exist_connector = []
        for connector in old_connectors:
            old_from_link_id = connector['from_link_id']
            old_to_link_id = connector['to_link_id']
            new_from_link_id = old_new_link_mapping[old_from_link_id][-1]  # 上游路段取新的link列表的最后一个
            new_to_link_id = old_new_link_mapping[old_to_link_id][0]  # 下游路段取新的link列表的第一个

            connector_name = f'{new_from_link_id}_{new_to_link_id}'
            if connector_name in exist_connector:
                continue
            self.netiface.createConnector3DWithPoints(new_from_link_id,
                                                      new_to_link_id,
                                                      [i[0] + 1 for i in connector['connector']],
                                                      [i[1] + 1 for i in connector['connector']],
                                                      connector['lanesWithPoints3'],
                                                      f"{new_from_link_id}-{new_to_link_id}"
                                                      )
            # 采用默认连接，防止出现连接段不平滑问题
            # netiface.createConnector(new_from_link_id, new_to_link_id, [i[0] + 1 for i in connector['connector']], [i[1] + 1 for i in connector['connector']], f"{new_from_link_id}-{new_to_link_id}")
            exist_connector.append(connector_name)


###############################################################################
# 简化路网文件所需函数


def get_section_childs(section_info: Dict, lengths: List, direction: str) -> List[Dict]:
    """
        处理某 Section 某方向的路网信息
        根据二维信息以及自定义配置，将section 进行再次处理，将宽度过窄的车道所在的路段打断，分配合适的连接段或者抹除车道
    Args:
        section_info:
        lengths: 因为分段时，以road作为最小单位，所以需要提供此 section 对应的长度断点序列
        direction: 沿参考线方向

    Returns:
        被处理后的 小路段列表
    """
    # 分为左右车道，同时过滤tess规则下不同的车道类型
    if direction == 'left':
        lane_ids = [lane_id for lane_id in section_info['lanes'].keys() if
                    lane_id > 0 and lane_id in section_info["tess_lane_ids"]]
    else:
        lane_ids = [lane_id for lane_id in section_info['lanes'].keys() if
                    lane_id < 0 and lane_id in section_info["tess_lane_ids"]]
    # 路段遍历，获取link段，处理异常点
    point_infos = []
    # 查找连接段
    for index, length in enumerate(lengths):
        point_info = {
            'lanes': {},
            'is_link': True,
        }
        for lane_id in lane_ids:
            lane_info = section_info['lanes'][lane_id]
            tess_lane_type = LANE_TYPE_MAPPING.get(lane_info['type'])
            if tess_lane_type not in WIDTH_LIMIT.keys() or lane_info['widths'][index] > \
                    WIDTH_LIMIT[tess_lane_type]['split']:
                point_info['lanes'][lane_id] = lane_info['type']  # 无宽度限制或者宽度足够，正常车道
            elif lane_info['widths'][index] > WIDTH_LIMIT[tess_lane_type]['join']:
                point_info['lanes'][lane_id] = lane_info['type']  # 宽度介于中间，作为连接段
                point_info['is_link'] = False
            # 否则，不加入车道列表
        point_infos.append(point_info)

    # 连续多个点的信息完全一致，可作为同一路段
    childs = []
    child_point = []
    start_index = None  # 分片时，start_index 为None 会视为0
    for index in range(len(lengths)):
        if len(child_point) == 1:
            start_index = index - 1
        point_info = point_infos[index]
        if index < POINT_REQUIRE:  # 首尾必须为link
            child_point.append(point_infos[0])
        elif len(lengths) - index - 1 < POINT_REQUIRE:
            child_point.append(point_infos[-1])
        elif not child_point:  # 原列表为空
            if point_info['is_link']:
                child_point.append(point_info)
            else:
                continue
        else:  # 原列表不为空
            if point_info == child_point[0]:
                child_point.append(point_info)
            elif len(child_point) >= POINT_REQUIRE:
                childs.append(
                    {
                        'start': start_index,
                        'end': index,
                        'lanes': set(child_point[0]['lanes'].keys()) & set(child_point[0]['lanes'].keys()),
                    }
                )
                child_point = []
            else:
                continue

    # 把最后一个存在的点序列导入, 最后一个应该以末尾点为准,但是如果此处包含了首&尾，应该取数据量的交集
    # 这样可能会丢失部分路段，所以在建立连接段时，必须确保 from_lane, to_lane 均存在
    childs.append(
        {
            'start': start_index,
            'end': len(lengths) - 1,
            'lanes': set(child_point[0]['lanes'].keys()) & set(child_point[0]['lanes'].keys())
        }
    )
    # lengths 只是断点序列，标记着与起始点的距离,反向用了同样的lengths
    # 得到link的列表,因为lane的点坐标与方向有关，所以此时的child已经根据方向排序
    return childs


def get_inter(string: str, roads_info: Dict) -> List:
    """
        根据 opendrive2lanelet 中定义的的车道名判断此车道的合法性，并返回相应的的 road_id,section_id,lane_id, -1
    Args:
        string: 车道名
        roads_info: 路段信息字典

    Returns:
        车道名是否合法，车道名的详细信息
    """
    inter_list = []
    is_true = True
    for i in string.split('.'):
        try:
            inter_list.append(int(i))
        except:
            inter_list.append(None)
            is_true = False

    # 检查 前后续路段 是否存在,不存在可以忽略
    if inter_list[0] not in roads_info.keys():
        is_true = False
    return [is_true, *inter_list]


def connect_childs(links: "Tessng.ILink", connector_mapping: Dict) -> None:
    """
        因为在section内对过窄车道进行了优化处理，所以可能将section切分成了多个link，需要为每个link在内部建立连接关系
    Args:
        links: section内 同方向子link 顺序列表
        connector_mapping: 全局的路段连接关系表

    Returns:
        无返回值，但路段连接关系表被扩充
    """
    for index in range(len(links) - 1):
        from_link_info = links[index]
        to_link_info = links[index + 1]
        from_link = from_link_info['link']
        to_link = to_link_info['link']
        if not (from_link and to_link and from_link_info['lane_ids'] and to_link_info['lane_ids']):
            continue

        actionTypeMapping = collections.defaultdict(lambda: {"from": set(), 'to': set()})
        for lane_id in from_link_info['lane_ids']:
            actionTypeMapping[from_link_info[lane_id].actionType()]['from'].add(lane_id)
        for lane_id in to_link_info['lane_ids']:
            actionTypeMapping[to_link_info[lane_id].actionType()]['to'].add(lane_id)

        connect_lanes = set()
        for actionType, lanes_info in actionTypeMapping.items():
            for lane_id in set(lanes_info['from'] | lanes_info['to']):
                # 车道原始编号相等，取原始编号对应的车道，否则，取临近车道(lane_ids 是有序的，所以临近车道永远偏向左边区)
                # 因为进行过自优化(人为打断与合并)，可能会导致前后失去连接关系，需要注意
                from_lane_id = min(lanes_info['from'], key=lambda x: abs(x - lane_id)) if lanes_info[
                    'from'] else None
                to_lane_id = min(lanes_info['to'], key=lambda x: abs(x - lane_id)) if lanes_info['to'] else None

                from_lane = from_link_info.get(from_lane_id)
                to_lane = to_link_info.get(to_lane_id)
                if from_lane and to_lane:
                    connect_lanes.add((from_lane.number() + 1, to_lane.number() + 1))

        if connect_lanes:
            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lFromLaneNumber'] += [i[0] for i in
                                                                                         connect_lanes]
            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lToLaneNumber'] += [i[1] for i in connect_lanes]
            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lanesWithPoints3'] += [None for _ in
                                                                                          connect_lanes]
            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['infos'] += []


# 计算两向量间的夹角
def cal_angle_of_vector(v0, v1, is_use_deg=True):
    dot_product = np.dot(v0, v1)
    v0_len = np.linalg.norm(v0)
    v1_len = np.linalg.norm(v1)
    try:
        angle_rad = np.arccos(dot_product / (v0_len * v1_len))
    except ZeroDivisionError as error:
        return None

    if is_use_deg:
        return np.rad2deg(angle_rad)
    return angle_rad


# 根据前后两点坐标绘制向量
def get_vector(start_point, end_point):
    # return np.array([end_point[index] - start_point[index] for index in range(3)])
    return np.array(end_point) - np.array(start_point)


# 根据角度限制整合中心点，减少绘制路网时的点数
def get_new_point_indexs(center_points, angle):
    new_point_indexs = [0, 1]  # 前两个点是基础
    for index, point in enumerate(center_points[2:-1]):
        real_index = index + 2  # 在 center_points 中的真实索引

        point_0 = center_points[new_point_indexs[-2]]
        point_1 = center_points[new_point_indexs[-1]]
        point_2 = point

        vector_1 = get_vector(point_0, point_1)
        vector_2 = get_vector(point_0, point_2)

        included_angle = cal_angle_of_vector(vector_1, vector_2)
        point_distance = np.sqrt(np.sum((np.array(point_2) - np.array(point_1)) ** 2))

        # 如果两向量夹角变化不大，抹除观测点(不能用连续的两个点比较,即 01 对比 12，容易对缓慢变化的路段判断错误)
        # print(included_angle, angle, point_distance >= max_length, included_angle > 0)
        # print(point_0, point_1, point_2, vector_1, vector_2)
        if included_angle is None or included_angle >= angle:
            new_point_indexs.append(real_index)
        # 如果观测点距离上一点过远，且存在角度差异，也不会抹除，不加这个可能会把微小的角度误差放大到明显的地步
        elif point_distance >= max_length and included_angle > 0:
        # elif point_distance >= max_length:  # 存在微小角度偏差的路段 可能会漫延很长
            new_point_indexs.append(real_index)
        else:
            # print('不添加')
            pass
    # 最后一个点需添加
    if len(center_points) - 1 > new_point_indexs[-1]:
        new_point_indexs.append(len(center_points) - 1)

    # TODO 尝试缩减第二个点
    def is_remove_index_1(center_points, new_point_indexs):
        if len(new_point_indexs) < 3:
            # 只有两个点，无法缩减
            return False
        point_0 = center_points[new_point_indexs[0]]
        point_1 = center_points[new_point_indexs[1]]
        point_2 = center_points[new_point_indexs[2]]

        vector_1 = get_vector(point_0, point_1)
        vector_2 = get_vector(point_0, point_2)

        included_angle = cal_angle_of_vector(vector_1, vector_2)

        # 距离取较小值
        point_distance = np.sqrt(np.sum((np.array(point_1) - np.array(point_0)) ** 2))

        # 如果两向量夹角变化不大，抹除观测点(不能用连续的两个点比较,即 01 对比 12，容易对缓慢变化的路段判断错误)
        if included_angle is None or included_angle >= angle:
            return False
        # 如果观测点距离上一点过远，且存在角度差异，也不会抹除，不加这个可能会把微小的角度误差放大到明显的地步
        elif point_distance >= max_length and included_angle > 0:
            # elif point_distance >= max_length:  # 存在微小角度偏差的路段 可能会蔓延很长
            return False
        else:
            return True

    if is_remove_index_1(center_points, new_point_indexs):
        del new_point_indexs[1]
    # print(f"point 缩减 {len(center_points)} to {len(new_point_indexs)}")
    return new_point_indexs

###############################################################################

# 简化路网文件
def simplify_tessng_file(file_path, angle, length):
    """
    简化 .tess 的路网，一个纯粹的方法，不依赖于tessng
    """
    default_angle = angle
    max_length = length

    message = "路网简化结果:\n\n"
    def dict_factory(cursor, row):
        d = {}
        for idx, col in enumerate(cursor.description):
            d[col[0]] = row[idx]
        return d

    conn = sqlite3.connect(file_path)  # 建立一个基于硬盘的数据库实例
    conn.row_factory = dict_factory

    cur = conn.cursor()  # 通过建立数据库游标对象，准备读写操作
    cur.execute("select * from link")  # 根据上表结构建立对应的表结构对象
    links = cur.fetchall()

    origin_point_counts = 0
    new_point_counts = 0
    for link in links:
        link_id = link['LinkID']
        cur.execute(
            f"select * from linkvertex join vertex on linkvertex.Vertexid=vertex.Vertexid where linkid={link_id} order by linkvertex.num")
        link_vertexs = cur.fetchall()
        center_points = [[point['X'], point['Y'], point['Z']] for point in link_vertexs]

        new_point_indexs = get_new_point_indexs(center_points, default_angle)
        if new_point_indexs[-1] != len(center_points) - 1:
            raise 1

        # 获取需要被删除的 顶点 ID
        delete_vertex_ids = []
        new_vertex_num_mapping = {}
        for index, vertex in enumerate(link_vertexs):
            if index not in new_point_indexs:
                delete_vertex_ids.append(vertex['VertexID'])
            else:
                new_vertex_num_mapping[vertex['VertexID']] = new_point_indexs.index(index) + 1

        # 判断 new_vertex_num_mapping 是否正确
        if max(new_vertex_num_mapping.values()) - min(new_vertex_num_mapping.values()) != len(
                new_vertex_num_mapping) - 1:
            raise 1

        # 先删除 link-vertex 关联表
        if len(delete_vertex_ids) == 1:
            sql = f"delete from linkvertex where linkid={link_id} and vertexid={delete_vertex_ids[0]}"
        else:
            sql = f"delete from linkvertex where linkid={link_id} and vertexid in {tuple(delete_vertex_ids)}"

        cur.execute(sql)
        # 更新 link-vertex 的 num
        for vertexid, num in new_vertex_num_mapping.items():
            sql = f'update linkvertex set num={num} where vertexid={vertexid}'
            cur.execute(sql)

        # 再删除顶点 vertex 表
        if len(delete_vertex_ids) == 1:
            sql = f"delete from vertex where vertexid={delete_vertex_ids[0]}"
        else:
            sql = f"delete from vertex where vertexid in {tuple(delete_vertex_ids)}"
        cur.execute(sql)

        # 更新路段的左右边界点
        try:
            point_count = set([len(json.loads(link[key])['data']) for key in
                               ['centerLinePointsJson', 'leftBreakPointsJson', 'rightBreakPointsJson']])
        except:
            continue

        if len(point_count) != 1:
            raise 1
        for key in ['centerLinePointsJson', 'leftBreakPointsJson', 'rightBreakPointsJson']:
            points = json.loads(link[key])['data']
            new_points = [point for index, point in enumerate(points) if index in new_point_indexs]
            sql = f'update link set {key}=:who where linkid={link_id}'
            cur.execute(sql, {"who": json.dumps({"data": new_points})})

        # 更新车道的左右边界点
        cur.execute(f"select * from lane where linkid={link_id}")
        lanes = cur.fetchall()

        for lane in lanes:
            lane_id = lane["LaneID"]
            point_count = set([len(json.loads(lane[key])['data']) for key in
                               ['centerLinePointsJson', 'leftBreakPointsJson', 'rightBreakPointsJson']])
            if len(point_count) != 1:
                raise 1

            for key in ['centerLinePointsJson', 'leftBreakPointsJson', 'rightBreakPointsJson']:
                points = json.loads(lane[key])['data']
                new_points = [point for index, point in enumerate(points) if index in new_point_indexs]
                sql = f'update lane set {key}=:who where laneid={lane_id}'
                cur.execute(sql, {"who": json.dumps({"data": new_points})})

        origin_point_counts += len(center_points)
        new_point_counts += len(new_point_indexs)

    message += f"原路段中可查询的点位共 {origin_point_counts} 个, 简化后剩余 {new_point_counts} 个, \n"

    cur.execute("select * from laneconnector")  # 根据上表结构建立对应的表结构对象
    laneconnectors = cur.fetchall()
    origin_point_counts = 0
    new_point_counts = 0
    for laneconnector in laneconnectors:
        lane = laneconnector
        points = json.loads(lane["centerLinePointsJson"])
        if not points:
            continue

        points = points['data']
        center_points = [[point['x'], point['y'], point['z']] for point in points]
        new_point_indexs = get_new_point_indexs(center_points, default_angle)

        for key in ['centerLinePointsJson', 'leftBreakPointsJson', 'rightBreakPointsJson']:
            points = json.loads(lane[key])['data']
            new_points = [point for index, point in enumerate(points) if index in new_point_indexs]
            sql = f'update laneconnector set {key}=:who where connID={laneconnector["connID"]} and startLaneid={laneconnector["StartLaneID"]} and endlaneid={laneconnector["EndLaneID"]}'
            cur.execute(sql, {"who": json.dumps({"data": new_points})})

        origin_point_counts += len(points)
        new_point_counts += len(new_points)

    message += f"原连接段中可查询的点位共 {origin_point_counts} 个, 简化后剩余 {new_point_counts} 个 \n"

    conn.commit()  # 保存提交，确保数据保存成功
    conn.close()  # 关闭与数据库的连接

    return message


from .tess2opendrive import tess2opendrive
from .tess2shape import tess2shape
from .tess2unity import tess2unity
from .tess2json import tess2json


def tess2other(netiface, params, mode):
    if mode == "opendrive":
        tess2opendrive.tess2opendrive(netiface, params)
    elif mode == "shape":
        tess2shape.tess2shape(netiface, params, mode="shape")
    elif mode == "geojson":
        tess2shape.tess2shape(netiface, params, mode="geojson")
    elif mode == "unity":
        tess2unity.tess2unity(netiface, params)
    elif mode == "json":
        tess2json.tess2json(netiface, params)
    else:
        raise Exception("No this export mode!")



from .opendrive2tess import opendrive2tess
from .shape2tess import shape2tess
from .osm2tess import osm2tess
from .excel2tess import excel2tess


def other2tess(netiface, params, mode):
    if mode == "opendrive":
        opendrive2tess.opendrive2tess(netiface, params)
    elif mode == "shape":
        shape2tess.shape2tess(netiface, params)
    elif mode == "osm":
        osm2tess.osm2tess(netiface, params)
    elif mode == "excel":
        excel2tess.excel2tess(netiface, params)
    else:
        raise Exception("No this import mode!")



# -*- coding: utf-8 -*-

from .roadPlanView import PlanView
from .roadLink import Link
from .roadLanes import Lanes
from .roadElevationProfile import ElevationProfile
from .roadLateralProfile import LateralProfile
from .junction import Junction


class Road:
    """ """

    def __init__(self):
        self._id = None
        self._name = None
        self._junction = None
        self._length = None

        self._header = None  # TODO
        self._link = Link()
        self._types = []
        self._planView = PlanView()
        self._elevationProfile = ElevationProfile()
        self._lateralProfile = LateralProfile()
        self._lanes = Lanes()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    @property
    def id(self):
        """ """
        return self._id

    @id.setter
    def id(self, value):
        """

        Args:
          value:

        Returns:

        """
        self._id = int(value)

    @property
    def name(self):
        """ """
        return self._name

    @name.setter
    def name(self, value):
        """

        Args:
          value:

        Returns:

        """
        self._name = str(value)

    @property
    def junction(self):
        """ """
        return self._junction

    @junction.setter
    def junction(self, value):
        """

        Args:
          value:

        Returns:

        """
        if not isinstance(value, Junction) and value is not None:
            raise TypeError("Property must be a Junction or NoneType")

        if value == -1:
            value = None

        self._junction = value

    @property
    def link(self):
        """ """
        return self._link

    @property
    def types(self):
        """ """
        return self._types

    @property
    def planView(self):
        """ """
        return self._planView

    @property
    def elevationProfile(self):
        """ """
        return self._elevationProfile

    @property
    def lateralProfile(self):
        """ """
        return self._lateralProfile

    @property
    def lanes(self):
        """ """
        return self._lanes



from .utils import config


def shape2tess(netiface, params):
    # params:
    # - folder_path
    # - is_use_center_line
    # - is_use_lon_and_lat
    # - laneFileName
    # - laneConnectorFileName
    # - proj_mode
    
    folder_path = params["folder_path"]
    is_use_lon_and_lat = params["is_use_lon_and_lat"]
    is_use_center_line = params["is_use_center_line"]
    laneFileName = params["laneFileName"]
    laneConnectorFileName = params["laneConnectorFileName"]
    proj_mode = params["proj_mode"]
    
    config.sceneScale = netiface.sceneScale()
    
    from .utils.get_data import get_data
    from .utils.create_network import create_network
    
    # 读取数据
    links_info, connector_info, other_info = get_data(folder_path, is_use_lon_and_lat, is_use_center_line, laneFileName, laneConnectorFileName, proj_mode)

    # import json
    # links_info_save = eval(str(dict(links_info)))
    # connector_info_save = {str(k):v for k,v in eval(str(dict(connector_info))).items()}
    # data = {"link": links_info_save, "connector": connector_info_save}
    # with open('data.json', 'w') as json_file:
    #     json.dump(data, json_file, indent=4)

    # 创建路段和连接段
    create_network(netiface, links_info, connector_info, other_info)


import os
import shapefile
# 核查shape文件的坐标类型
def check_shapefile(folder_path, laneFileName, is_use_lon_and_lat):
    # 文件路径
    filePath_shp_lane = os.path.join(folder_path, f"{laneFileName}.shp")

    # 读取文件
    try:
        all_data_shp_lane = shapefile.Reader(filePath_shp_lane, encoding='utf-8').shapes()
    except:
        all_data_shp_lane = shapefile.Reader(filePath_shp_lane, encoding='gbk').shapes()

    # 坐标范围
    x_list, y_list = [], []
    for lane_data_shp in all_data_shp_lane:
        x_list.extend([point[0] for point in lane_data_shp.points])
        y_list.extend([point[1] for point in lane_data_shp.points])

    x_diff = max(x_list) - min(x_list)
    y_diff = max(y_list) - min(y_list)

    # 经纬度坐标
    if is_use_lon_and_lat:
        # 一定不是
        if min(x_list) < -180 or max(x_list) > 180 or min(y_list) < -90 or max(y_list) > 90:
            is_ok = False
            prompt_information = "判断为笛卡尔坐标，请检查！"
        # 大概率不是
        elif x_diff > 2 or y_diff > 2:
            is_ok = False
            prompt_information = "极大概率为笛卡尔坐标，请检查！"
        # 大概率是
        else:
            is_ok = True
            prompt_information = ""
    # 笛卡尔坐标
    else:
        # 很可能不是
        if x_diff < 1 and y_diff < 1:
            is_ok = False
            prompt_information = "极大概率为经纬度坐标，请检查！"
        else:
            is_ok = True
            prompt_information = ""

    return is_ok, prompt_information


from .utils import config


def tess2opendrive(netiface, params):
    # params:
    # - file_path
    # - proj_string
    
    # 文件保存路径
    file_path = params["file_path"]
    
    # 场景比例尺
    config.sceneScale = netiface.sceneScale()
    
    # 导入计算模块
    from .utils.get_data import get_data
    
    # 获取数据
    xml_pretty_str = get_data(netiface, params["proj_string"])
    
    # 写入数据
    with open(file_path, 'w', encoding='utf-8') as f:
        f.write(xml_pretty_str)


from PySide2.QtGui import QVector3D
from PySide2.QtCore import QPointF
from .config import sceneScale


def m2p(x):
    return x / sceneScale

def get_coo_list(vertices):
    list1 = []
    x_move, y_move = 0, 0
    for index in range(0, len(vertices), 1):
        vertice = vertices[index]
        if len(vertice) == 3:
            list1.append(QVector3D(m2p((vertice[0] + x_move)), m2p(-(vertice[1]) + y_move), m2p(vertice[2])))
        elif len(vertice) == 2:
            list1.append(QPointF(m2p((vertice[0] + x_move)), m2p(-(vertice[1]) + y_move)))
    if len(list1) < 2:
        raise 3
    return list1



import collections
import copy
import json
import sqlite3
import traceback
import numpy as np
from PySide2.QtGui import QVector3D
from typing import List, Dict
from .config import LANE_TYPE_MAPPING, MIN_CONNECTOR_LENGTH, is_center, default_angle, simplify_network_force
from .convert_utils import convert_opendrive, convert_roads_info, convert_lanes_info
from .functions import p2m, m2p
from .functions import get_inter, get_section_childs, connect_childs, get_new_point_indexs
from ..opendrive2lanelet.opendriveparser.elements.opendrive import OpenDrive
from ..opendrive2lanelet.opendriveparser.elements.roadLanes import Lane
try:
    from .....Tessng.MyMenu import ProgressDialog as pgd
except:
    class pgd: progress = lambda values, text="": values


class Section:
    def __init__(self, road_id, section_id, lane_ids: list):
        self.road_id = road_id
        self.id = section_id
        self._left_link = None
        self._right_link = None
        self.lane_ids = list(lane_ids or [])
        # 左右来向的车道id分别为正负， 需要根据tess的规则进行排序
        self.left_lane_ids = sorted(filter(lambda i: i > 0, self.lane_ids, ), reverse=True)
        self.right_lane_ids = sorted(filter(lambda i: i < 0, self.lane_ids, ), reverse=False)
        self.lane_mapping = {}

    @property
    def left_link(self):
        return self._left_link

    @left_link.setter
    def left_link(self, obj):
        for link_info in obj:
            link = link_info['link']
            if not link:
                continue
            lane_ids = link_info['lane_ids']
            for index, lane in enumerate(link.lanes()):
                link_info[lane_ids[index]] = lane
        self._left_link = obj

    @property
    def right_link(self):
        return self._right_link

    @right_link.setter
    def right_link(self, obj):
        for link_info in obj:
            link = link_info['link']
            # 路段创建失败时，link 为 None
            if not link:
                continue
            lane_ids = link_info['lane_ids']
            for index, lane in enumerate(link.lanes()):
                link_info[lane_ids[index]] = lane
        self._right_link = obj

    def tess_lane(self, lane_id, type):
        try:
            attr = self.left_link if lane_id > 0 else self.right_link
            link_index = -1 if type == 'from' else 0
            return attr[link_index].get(lane_id)
        except:
            return None

    def tess_link(self, lane_id, type):
        try:
            attr = self.left_link if lane_id > 0 else self.right_link
            link_index = -1 if type == 'from' else 0
            return attr[link_index]['link']
        except:
            return None


class Road:
    def __init__(self, road_id):
        self.id = road_id
        self.sections = []

    def section(self, section_id: int = None):
        if section_id is None:
            return self.sections
        else:
            for section in self.sections:
                if section.id == section_id:
                    return section

    def section_append(self, section: Section):
        self.sections.append(section)
        self.sections.sort(key=lambda i: i.id)


class Network:
    def __init__(self, opendrive: OpenDrive):
        """
            TessNg 路网对象初始化
        Args:
            opendrive: 
        """
        self.opendrive = opendrive
        self.network_info = None
        self.xy_move = (0, 0)
        self.size = (300, 600)
        self.step = None

    def extract_network_info(self, step: int = None, filters: List[str] = None, context: Dict = None):
        """
            借助 opendrive2lanelet 开源库提取opendrive基础信息，包括路段/车道三维信息以及车道间的连接关系
        Args:
            step: 提取信息时的精度
            filters: 过滤的opendrive车道类型
            context: 上下文信息

        Returns:

        """
        step = step or 1
        self.step = step
        filters = filters or Lane.laneTypes
        opendrive = self.opendrive
        # 头信息
        header_info = {
            "date": opendrive.header.date,
            "geo_reference": opendrive.header.geo_reference,
        }

        # 参考线信息解析
        roads_info = convert_roads_info(opendrive, step, filters)

        # 车道点位序列不再独立计算，采用 road info 中参考线的点位
        # 车道信息解析，这一步最消耗时间，允许传入进度条
        scenario = convert_opendrive(opendrive, filters, roads_info, context)
        lanes_info = convert_lanes_info(opendrive, scenario, roads_info)

        network_info = {
            "header_info": header_info,
            "roads_info": roads_info,
            "lanes_info": lanes_info,
        }
        # print("network_info:", network_info)
        return network_info

    def convert_network(self, step=None, filters=None, context=None):
        """
        提取opendrive路网基础信息

        Args:
            step:
            filters:
            context:

        Returns:

        """
        try:
            self.network_info = self.extract_network_info(step, filters, context)
            roads_info = self.network_info["roads_info"]
            lanes_info = self.network_info["lanes_info"]

            xy_limit = None
            for road_id, road_info in pgd.progress(roads_info.items(), '数据处理中（6/13）'):
                # 记录 坐标点的极值 (取左右point列表无区别，只是计算方向不同)
                for section_id, points in road_info['road_points'].items():
                    for point in points['right_points']:
                        position = point['position']
                        if xy_limit is None:  # x1,x2,y1,y2
                            xy_limit = [position[0], position[0], position[1], position[1]]
                        else:
                            xy_limit[0] = min(xy_limit[0], position[0])
                            xy_limit[1] = max(xy_limit[1], position[0])
                            xy_limit[2] = min(xy_limit[2], position[1])
                            xy_limit[3] = max(xy_limit[3], position[1])

            self.xy_move = (- sum(xy_limit[:2]) / 2, - sum(xy_limit[2:]) / 2) if xy_limit else (0, 0)
            self.size = (max(abs(xy_limit[0]), abs(xy_limit[1])) * 2, max(abs(xy_limit[2]), abs(xy_limit[3])) * 2)
            # print(f"路网移动参数: {self.xy_move}")

            for lane_name, lane_info in pgd.progress(lanes_info.items(), '数据处理中（7/13）'):
                if not lane_info:  # 此车道只是文件中某车道的前置或者后置车道，仅仅被提及，是空信息，跳过
                    continue
                road_id = lane_info['road_id']
                section_id = lane_info['section_id']
                lane_id = lane_info['lane_id']
                if road_id not in roads_info.keys():
                    continue

                # 添加默认属性
                roads_info[road_id].setdefault('sections', {})
                roads_info[road_id]['sections'].setdefault(section_id, {})
                roads_info[road_id]['sections'][section_id].setdefault('lanes', {})
                roads_info[road_id]['sections'][section_id]["lanes"][lane_id] = lane_info

        except Exception as e:
            traceback.print_tb(e.__traceback__)
            print("失败了")

    def create_network(self, tess_lane_types, netiface=None):
        """
        在tessng中绘制路网
        Args:
            tess_lane_types: config 文件中定义了车道类型映射，这是其中的值集合，只有被选中的的tessng车道类型才会被创建
            netiface: TESSNG 路网子接口
        """
        # move
        if is_center:
            x_move = round(self.xy_move[0], 3)
            y_move = round(self.xy_move[1], 3)
        else:
            x_move, y_move = 0, 0
        other_info = {"move_distance": {"x_move": float(x_move), "y_move": float(y_move)}}
        # 记录属性，如：路网来源
        netiface.setNetAttrs("OpenDrive Network", otherAttrsJson=other_info)
        print("Network attrs: ", netiface.netAttrs().otherAttrs())
        # 设置场景显示区域
        netiface.setSceneSize(*self.size)

        # 会改变数据结构，可能在重复创建时有影响，所以创建新的数据备份
        for i in pgd.progress([1, 2], '数据处理中（8/13）'):
            if i == 1:
                roads_info = copy.deepcopy(self.network_info["roads_info"])
            else:
                lanes_info = copy.deepcopy(self.network_info["lanes_info"])

        # 对于宽度过窄的车道所在路段进行打断
        for road_id, road_info in pgd.progress(roads_info.items(), '数据处理中（9/13）'):
            for section_id, section_info in road_info.get('sections', {}).items():
                section_info['tess_lane_ids'] = [lane_id for lane_id, lane_info in section_info['lanes'].items() if
                                                 LANE_TYPE_MAPPING.get(lane_info['type']) in tess_lane_types]

                lengths = road_info['road_points'][section_id]['lengths']
                section_info['left_childs'] = get_section_childs(section_info, lengths, 'left')
                section_info['right_childs'] = get_section_childs(section_info, lengths, 'right')

        # 初始化 错误信息 及 连接关系表
        error_junction = []
        connector_mapping = collections.defaultdict(lambda: {
            'lFromLaneNumber': [],
            'lToLaneNumber': [],
            'lanesWithPoints3': [],
            'infos': [],
        })
        # 创建路段并记录section内部的连接关系
        road_mapping = self.create_links(netiface, roads_info, connector_mapping, error_junction)
        # 记录路段间及section间的连接关系
        self.convert_link_connect(roads_info, lanes_info, connector_mapping, road_mapping, error_junction)
        # 记录交叉口的连接关系
        self.convert_junction(roads_info, lanes_info, connector_mapping, road_mapping, error_junction)
        # 实现所有的连接关系
        self.create_connects(netiface, connector_mapping)
        return error_junction

    def create_links(self, netiface, roads_info: Dict, connector_mapping: Dict,
                     error_junction: List) -> Dict:
        """
           在 TessNg 中创建 opendrive 所有的基础路段(Link),如果 子section被打断，同时将连接关系记录在 connector_mapping 中
        Args:
            netiface: TessNg 路网子接口
            roads_info: 路段详情字典
            connector_mapping: 全局的连接关系字典
            error_junction: 全局的错误信息列表

        Returns:
            opendrive 路段ID 与 TessNg Link 对象映射关系

        """
        road_mapping = dict()
        # 先行创建所有的基本路段
        for road_id, road_info in pgd.progress(roads_info.items(), '路段创建中（10/13）'):
            if not road_info['junction_id'] is None:
                continue
            tess_road = Road(road_id)
            for section_id, section_info in road_info.get('sections', {}).items():
                lengths = road_info['road_points'][section_id]['lengths']

                # section里的多段link已经根据方向重新排序
                tess_section = Section(road_id, section_id, section_info['tess_lane_ids'])
                tess_road.sections.append(tess_section)

                if not section_info['tess_lane_ids']:
                    continue
                for direction in ['left', 'right']:
                    # 判断此方向的路段中是否存在车道
                    is_exist = max(section_info['tess_lane_ids']) > 0 if direction == 'left' else min(
                        section_info['tess_lane_ids']) < 0
                    if not is_exist:
                        continue

                    points = road_info['road_points'][section_id][f'{direction}_points']
                    # 对section分段
                    section_links = []
                    # 记录了所有的路段(断点)
                    childs = section_info[f'{direction}_childs']
                    # 右车道id为负，越小的越先在tess中创建，左车道id为正，越大的越先创建
                    reverse = True if direction == 'left' else False
                    for index in range(len(childs)):
                        child = childs[index]
                        # 步长过大，可能会导致在分段时 child 只包含了一个点
                        start_index, end_index = child['start'], child['end'] + 1
                        land_ids = sorted(child['lanes'], reverse=reverse)  # 列表内多点的的数据是一样的，取第一个即可

                        # 是否需要简化路网
                        if not simplify_network_force:
                            lCenterLinePoint = self.get_coo_list(
                                [point["position"] for point in points][start_index:end_index])
                            lanesWithPoints = [
                                {
                                    'left': self.get_coo_list(
                                        road_info['sections'][section_id]["lanes"][lane_id]['left_vertices'][
                                        start_index:end_index]),
                                    'center': self.get_coo_list(
                                        road_info['sections'][section_id]["lanes"][lane_id]['center_vertices'][
                                        start_index:end_index]),
                                    'right': self.get_coo_list(
                                        road_info['sections'][section_id]["lanes"][lane_id]['right_vertices'][
                                        start_index:end_index]),
                                }
                                for lane_id in land_ids
                            ]
                        else:
                            center_points = [point["position"] for point in points][start_index:end_index]
                            new_point_indexs = get_new_point_indexs(center_points, default_angle)
                            new_center_points = [center_points[index] for index in new_point_indexs]
                            lCenterLinePoint = self.get_coo_list(new_center_points)
                            lanesWithPoints = []
                            for lane_id in land_ids:
                                lane_left_points = road_info['sections'][section_id]["lanes"][lane_id]['left_vertices'][start_index:end_index]
                                lane_center_points = road_info['sections'][section_id]["lanes"][lane_id]['center_vertices'][start_index:end_index]
                                lane_right_points = road_info['sections'][section_id]["lanes"][lane_id]['right_vertices'][start_index:end_index]

                                new_lane_left_points = [lane_left_points[index] for index in new_point_indexs]
                                new_lane_center_points = [lane_center_points[index] for index in new_point_indexs]
                                new_lane_right_points = [lane_right_points[index] for index in new_point_indexs]
                                lanesWithPoints.append(
                                    {
                                        'left': self.get_coo_list(new_lane_left_points),
                                        'center': self.get_coo_list(new_lane_center_points),
                                        'right': self.get_coo_list(new_lane_right_points),
                                    }
                                )
                            # print(len(center_points), len(new_center_points))

                        lLaneType = [LANE_TYPE_MAPPING[section_info['lanes'][lane_id]['type']] for lane_id in land_ids]
                        lAttr = [{} for _ in land_ids]
                        link_name = f"{road_id}_{section_id}_{index}_{direction}"

                        link_obj = netiface.createLink3DWithLanePointsAndAttrs(lCenterLinePoint, lanesWithPoints,
                                                                               lLaneType, lAttr, link_name)
                        # link_obj = netiface.createLink3DWithLaneWidth(lCenterLinePoint, [m2p(i+1) for i in range(len(lAttr))], 'link_name')

                        # link_obj 可能为None, 为什么会发生这种情况
                        if not link_obj:
                            error_junction.append(
                                {
                                    "roadName": f"{road_id}_{section_id}_{index}_{direction}",
                                    "centerLine": [point["position"] for point in points][start_index:end_index],
                                    "landIds": land_ids,
                                    "message": "路段创建失败",
                                }
                            )
                        link_info = {
                            'link': link_obj,
                            'lane_ids': land_ids
                        }
                        section_links.append(link_info)
                    tess_section.__setattr__(f"{direction}_link", section_links)
                    connect_childs(getattr(tess_section, f"{direction}_link"), connector_mapping)
            road_mapping[road_id] = tess_road

        return road_mapping

    def convert_link_connect(self, roads_info, lanes_info, connector_mapping, road_mapping, error_junction) -> None:
        """
            获取所有 opendrive 中正常路段及section(不含section内部)间的连接关系并记录
        Args:
            roads_info:
            lanes_info:
            connector_mapping:
            road_mapping:
            error_junction:

        Returns:
            无返回值，所有的连接关系被记录在全局变量 connector_mapping 中
        """
        # 累计所有的路段间的连接段
        link_road_ids = [road_id for road_id, road_info in roads_info.items() if road_info['junction_id'] is None]
        junction_road_ids = [road_id for road_id, road_info in roads_info.items() if
                             not road_info['junction_id'] is None]

        for road_id in pgd.progress(link_road_ids, '连接关系记录中（11/13）'):
            road_info = roads_info[road_id]
            # lane_sections 保存基本信息，sections 保存详情
            for section_id, section_info in road_info.get('sections', {}).items():
                # 路段间的连接段只向后连接,本身作为前路段(向前也一样，会重复一次)
                for lane_id in section_info["tess_lane_ids"]:
                    lane_info = section_info['lanes'][lane_id]
                    predecessor_id = lane_info['name']
                    # 为了和交叉口保持一致，重新获取一次相关信息
                    is_true, from_road_id, from_section_id, from_lane_id, _ = get_inter(predecessor_id, roads_info)

                    # 部分车道的连接关系可能是'2.3.None.-1', 需要清除（上一车道宽度归零，不会连接到下一车道）
                    # 同时 predecessor_id 的 from_road 可能并不在路网中
                    if not (is_true and from_road_id in road_mapping.keys()):
                        continue

                    from_section = road_mapping[from_road_id].section(from_section_id)
                    from_link = from_section and from_section.tess_link(from_lane_id, 'from')
                    from_lane = from_section and from_section.tess_lane(from_lane_id, 'from')

                    for successor_id in lane_info["successor_ids"]:
                        is_true, to_road_id, to_section_id, to_lane_id, _ = get_inter(successor_id, roads_info)
                        if not (is_true and to_road_id in road_mapping.keys()):
                            continue

                        if to_road_id not in link_road_ids:
                            continue

                        # 查看 section 为什么为 None
                        to_section = road_mapping[to_road_id].section(to_section_id)
                        to_link = to_section and to_section.tess_link(to_lane_id, 'to')
                        to_lane = to_section and to_section.tess_lane(to_lane_id, 'to')

                        # 多种原因会出现这种情况,1.如果车道在此处宽度归零，是不会连接到下一车道的 2.步长过大导致取点时进行了交集
                        if not (from_link and from_lane and to_link and to_lane):
                            continue

                        if from_lane.actionType() != to_lane.actionType():
                            error_junction.append(
                                {
                                    "from_lane_id": f"{from_road_id}, {from_section_id}, {from_lane_id}",
                                    "from_lane_type": lanes_info[predecessor_id]['type'],
                                    "to_lane_id": f"{to_road_id}, {to_section_id}, {to_lane_id}",
                                    "to_lane_type": lanes_info[successor_id]['type'],
                                    "message": f"连接段前后车道类型不同",
                                }
                            )
                            continue

                        connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lFromLaneNumber'].append(
                            from_lane.number() + 1)
                        connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lToLaneNumber'].append(
                            to_lane.number() + 1)

                        # 路段间连接需要自动连接
                        connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lanesWithPoints3'].append(None)
                        connector_mapping[f"{from_link.id()}-{to_link.id()}"]['infos'].append(
                            {
                                "predecessor_id": predecessor_id,
                                "successor_id": successor_id,
                                'junction': False,
                            }
                        )

    def convert_junction(self, roads_info, lanes_info, connector_mapping, road_mapping, error_junction):
        """
            提取交叉口的车道三维信息及连接关系，将其记录在 connector_mapping 中
        Args:
            roads_info:
            lanes_info:
            connector_mapping:
            road_mapping:
            error_junction:

        Returns:
            无返回值，交叉口的连接关系被记录在全局变量 connector_mapping 中
        """

        def get_predecessor_ids_by_link(new_predecessor_ids, predecessor_ids):
            """
                因为 opendrive 中交叉口也是road，也会有多个section，但是TessNg只会将其视作连接段
                所以，我们需要递归获取上下游连接段，直至获取到在正常路段(road非junction)的上下游车道
            """
            for predecessor_id in predecessor_ids:
                if predecessor_id in lanes_info.keys() and roads_info.get(lanes_info[predecessor_id]['road_id'],
                                                                          {}).get(
                    'junction_id') is not None:  # 前置路线为交叉口，继续向前遍历
                    get_predecessor_ids_by_link(new_predecessor_ids, lanes_info[predecessor_id]['predecessor_ids'])
                else:
                    new_predecessor_ids.append(predecessor_id)

        def get_successor_ids_by_link(new_successor_ids, successor_ids):
            for successor_id in successor_ids:
                if successor_id in lanes_info.keys() and roads_info.get(lanes_info[successor_id]['road_id'], {}).get(
                        'junction_id') is not None:  # 后置路线为交叉口，继续向后遍历
                    get_successor_ids_by_link(new_successor_ids, lanes_info[successor_id]['successor_ids'])
                else:
                    new_successor_ids.append(successor_id)

        link_road_ids = [road_id for road_id, road_info in roads_info.items() if road_info['junction_id'] is None]
        junction_road_ids = [road_id for road_id, road_info in roads_info.items() if
                             not road_info['junction_id'] is None]
        #  递归重分配交叉口的连接
        for road_id in junction_road_ids:
            road_info = roads_info[road_id]
            for section_id, section_info in road_info.get('sections', {}).items():
                # 获取路口的所有连接关系
                for lane_id in section_info['tess_lane_ids']:
                    lane_info = section_info["lanes"][lane_id]
                    new_predecessor_ids = []
                    new_successor_ids = []
                    get_predecessor_ids_by_link(new_predecessor_ids, lane_info['predecessor_ids'])
                    get_successor_ids_by_link(new_successor_ids, lane_info['successor_ids'])
                    lane_info['is_junction_change'] = new_predecessor_ids != lane_info[
                        'predecessor_ids'] or new_successor_ids != lane_info['successor_ids']
                    lane_info['predecessor_ids'] = new_predecessor_ids
                    lane_info['successor_ids'] = new_successor_ids

        # 仅交叉口
        for road_id in pgd.progress(junction_road_ids, '连接关系记录中（12/13）'):
            road_info = roads_info[road_id]
            for section_id, section_info in road_info.get('sections', {}).items():
                # 获取路口的所有连接关系
                for lane_id in section_info['tess_lane_ids']:
                    lane_info = section_info["lanes"][lane_id]

                    # 前后 lane 如果也在 junction 中，需要继续探索--> 重塑前后连接关系
                    for predecessor_id in lane_info['predecessor_ids']:
                        is_true, from_road_id, from_section_id, from_lane_id, _ = get_inter(predecessor_id, roads_info)
                        if not (is_true and from_road_id in road_mapping.keys()):
                            continue

                        from_section = road_mapping[from_road_id].section(from_section_id)
                        from_link = from_section and from_section.tess_link(from_lane_id, 'from')
                        from_lane = from_section and from_section.tess_lane(from_lane_id, 'from')

                        for successor_id in lane_info["successor_ids"]:
                            is_true, to_road_id, to_section_id, to_lane_id, _ = get_inter(successor_id, roads_info)
                            if not (is_true and to_road_id in road_mapping.keys()):  # 部分车道的连接关系可能是'2.3.None.-1'，需要清除
                                continue

                            to_section = road_mapping[to_road_id].section(to_section_id)
                            to_link = to_section and to_section.tess_link(to_lane_id, 'to')
                            to_lane = to_section and to_section.tess_lane(to_lane_id, 'to')

                            if not (from_link and from_lane and to_link and to_lane):
                                continue
                            # 检查车道类型是否异常
                            if from_lane.actionType() != to_lane.actionType():
                                error_junction.append(
                                    {
                                        "from_lane_id": f"{from_road_id}, {from_section_id}, {from_lane_id}",
                                        "from_lane_type": lanes_info[predecessor_id]['type'],
                                        "to_lane_id": f"{to_road_id}, {to_section_id}, {to_lane_id}",
                                        "to_lane_type": lanes_info[successor_id]['type'],
                                        "message": f"连接段前后车道类型不同",
                                    }
                                )
                                continue

                            # TODO 交叉口可能产生自连接，记录并跳过（连接发生在同一路段时，必须保证是同一个section）
                            if from_road_id == to_road_id and from_section_id != to_section_id:
                                error_junction.append(
                                    {
                                        "from_link_id": from_link.id(),
                                        "from_lane_number": from_lane.number() + 1,
                                        "from_lane_type": from_lane.actionType(),
                                        "to_link_id": to_link.id(),
                                        "to_lane_number": to_lane.number() + 1,
                                        "to_lane_type": from_lane.actionType(),
                                        "message": "车道连接信息错误",
                                    }
                                )
                                continue

                            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lFromLaneNumber'].append(
                                from_lane.number() + 1)
                            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lToLaneNumber'].append(
                                to_lane.number() + 1)

                            # 用前后车道的首尾坐标替换原有首尾坐标
                            if not lane_info.get('is_junction_change'):
                                center_connector_vertices = lanes_info[predecessor_id]['center_vertices'][-1:] + \
                                                            lane_info['center_vertices'][1:-1] + \
                                                            lanes_info[successor_id]['center_vertices'][:1]
                                left_connector_vertices = lanes_info[predecessor_id]['left_vertices'][-1:] + \
                                                          lane_info['left_vertices'][1:-1] + \
                                                          lanes_info[successor_id]['left_vertices'][:1]
                                right_connector_vertices = lanes_info[predecessor_id]['right_vertices'][-1:] + \
                                                           lane_info['right_vertices'][1:-1] + \
                                                           lanes_info[successor_id]['right_vertices'][:1]
                                connector_vertices = {
                                    "center": self.get_coo_list(center_connector_vertices),
                                    "left": self.get_coo_list(left_connector_vertices),
                                    "right": self.get_coo_list(right_connector_vertices),
                                }
                            else:
                                # TODO junction 中连接关系被重置，是否需要采用原有的连接段参数？
                                connector_vertices = None
                            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['lanesWithPoints3'].append(
                                connector_vertices
                            )

                            connector_mapping[f"{from_link.id()}-{to_link.id()}"]['infos'].append(
                                {
                                    "predecessor_id": predecessor_id,
                                    "successor_id": successor_id,
                                    "lane_id": lane_info["name"],
                                    "connector_vertices": connector_vertices,
                                    'junction': True,
                                    "from_link": from_link,
                                    "to_link": to_link,
                                }
                            )

    def create_connects(self, netiface: "TessInterface.netInterface", connector_mapping: Dict) -> None:
        """
            在TessNg中创建所有的连接段
            包括路段间，section间，section内部，交叉口
        """
        # 创建所有的连接关系
        for link_id, link_info in pgd.progress(connector_mapping.items(), '连接段创建中（13/13）'):
            from_link_id = int(link_id.split('-')[0])
            to_link_id = int(link_id.split('-')[1])
            lFromLaneNumber = link_info['lFromLaneNumber']
            lToLaneNumber = link_info['lToLaneNumber']
            lanesWithPoints3 = link_info['lanesWithPoints3']
            if not len(lanesWithPoints3) == len(lFromLaneNumber):
                raise

            # lFromLaneNumber,lToLaneNumber,lanesWithPoints3 需要去重
            new_connect_mapping = dict()
            max_connector_length = 0
            for index in range(len(lFromLaneNumber)):
                laneWithPoints3 = lanesWithPoints3[index]
                new_connect_mapping[(lFromLaneNumber[index], lToLaneNumber[index])] = laneWithPoints3
                if laneWithPoints3:
                    points = laneWithPoints3['center']
                    connector_length = np.linalg.norm(
                        np.array([p2m(points[0].x()), - p2m(points[0].y()), p2m(points[0].z())]) - np.array(
                            [p2m(points[1].x()), - p2m(points[1].y()), p2m(points[1].z())]))
                    max_connector_length = max(max_connector_length, connector_length)

            new_connect_list = [[key[0], key[1], value] for key, value in new_connect_mapping.items()]
            lFromLaneNumber, lToLaneNumber, lanesWithPoints3 = [i[0] for i in new_connect_list], [i[1] for i in
                                                                                                  new_connect_list], [
                                                                   i[2] for i in new_connect_list]

            # 源数据建立连接, 如果 MIN_CONNECTOR_LENGTH 存在，则要求 连接车道段的最大值大于 MIN_CONNECTOR_LENGTH
            if all(lanesWithPoints3) and (not MIN_CONNECTOR_LENGTH or max_connector_length >= MIN_CONNECTOR_LENGTH):
                netiface.createConnector3DWithPoints(from_link_id, to_link_id, lFromLaneNumber, lToLaneNumber,
                                                     lanesWithPoints3, f"{from_link_id}-{to_link_id}")
                # TESS 自动计算，建立连接
            else:
                netiface.createConnector(from_link_id, to_link_id, lFromLaneNumber, lToLaneNumber,
                                         f"{from_link_id}-{to_link_id}")


    def get_coo_list(self, vertices: List) -> List[QVector3D]:
        """
            将米制的三维点位坐标转换为QT的点位，如果要进行平移/旋转也是在这里进行
        """
        if is_center:
            x_move, y_move = self.xy_move
        else:
            x_move, y_move = 0, 0
        temp_list = []
        for vertice in vertices:
            temp_list.append(QVector3D(m2p((vertice[0] + x_move)), m2p(-(vertice[1] + y_move)), m2p(vertice[2])))
        return temp_list

    @classmethod
    def simplify_tessng_file(self, file_path):
        """
        简化 .tess 的路网，一个纯粹的方法，不依赖于tessng
        """
        message = "路网简化结果:\n"
        def dict_factory(cursor, row):
            d = {}
            for idx, col in enumerate(cursor.description):
                d[col[0]] = row[idx]
            return d

        conn = sqlite3.connect(file_path)  # 建立一个基于硬盘的数据库实例
        conn.row_factory = dict_factory

        cur = conn.cursor()  # 通过建立数据库游标对象，准备读写操作
        cur.execute("select * from link")  # 根据上表结构建立对应的表结构对象
        links = cur.fetchall()

        origin_point_counts = 0
        new_point_counts = 0
        for link in links:
            link_id = link['LinkID']
            cur.execute(
                f"select * from linkvertex join vertex on linkvertex.Vertexid=vertex.Vertexid where linkid={link_id} order by linkvertex.num")
            link_vertexs = cur.fetchall()
            center_points = [[point['X'], point['Y'], point['Z']] for point in link_vertexs]

            new_point_indexs = get_new_point_indexs(center_points, default_angle)
            if new_point_indexs[-1] != len(center_points) - 1:
                raise 1

            # 获取需要被删除的 顶点 ID
            delete_vertex_ids = []
            new_vertex_num_mapping = {}
            for index, vertex in enumerate(link_vertexs):
                if index not in new_point_indexs:
                    delete_vertex_ids.append(vertex['VertexID'])
                else:
                    new_vertex_num_mapping[vertex['VertexID']] = new_point_indexs.index(index) + 1

            # 判断 new_vertex_num_mapping 是否正确
            if max(new_vertex_num_mapping.values()) - min(new_vertex_num_mapping.values()) != len(
                    new_vertex_num_mapping) - 1:
                raise 1

            # 先删除 link-vertex 关联表
            if len(delete_vertex_ids) == 1:
                sql = f"delete from linkvertex where linkid={link_id} and vertexid={delete_vertex_ids[0]}"
            else:
                sql = f"delete from linkvertex where linkid={link_id} and vertexid in {tuple(delete_vertex_ids)}"

            cur.execute(sql)
            # 更新 link-vertex 的 num
            for vertexid, num in new_vertex_num_mapping.items():
                sql = f'update linkvertex set num={num} where vertexid={vertexid}'
                cur.execute(sql)

            # 再删除顶点 vertex 表
            if len(delete_vertex_ids) == 1:
                sql = f"delete from vertex where vertexid={delete_vertex_ids[0]}"
            else:
                sql = f"delete from vertex where vertexid in {tuple(delete_vertex_ids)}"
            cur.execute(sql)

            # 更新路段的左右边界点
            try:
                point_count = set([len(json.loads(link[key])['data']) for key in
                                   ['centerLinePointsJson', 'leftBreakPointsJson', 'rightBreakPointsJson']])
            except:
                continue

            if len(point_count) != 1:
                raise 1
            for key in ['centerLinePointsJson', 'leftBreakPointsJson', 'rightBreakPointsJson']:
                points = json.loads(link[key])['data']
                new_points = [point for index, point in enumerate(points) if index in new_point_indexs]
                sql = f'update link set {key}=:who where linkid={link_id}'
                cur.execute(sql, {"who": json.dumps({"data": new_points})})

            # 更新车道的左右边界点
            cur.execute(f"select * from lane where linkid={link_id}")
            lanes = cur.fetchall()

            for lane in lanes:
                lane_id = lane["LaneID"]
                point_count = set([len(json.loads(lane[key])['data']) for key in
                                   ['centerLinePointsJson', 'leftBreakPointsJson', 'rightBreakPointsJson']])
                if len(point_count) != 1:
                    raise 1

                for key in ['centerLinePointsJson', 'leftBreakPointsJson', 'rightBreakPointsJson']:
                    points = json.loads(lane[key])['data']
                    new_points = [point for index, point in enumerate(points) if index in new_point_indexs]
                    sql = f'update lane set {key}=:who where laneid={lane_id}'
                    cur.execute(sql, {"who": json.dumps({"data": new_points})})

            origin_point_counts += len(center_points)
            new_point_counts += len(new_point_indexs)

        message += f"原路段中可查询的点位共 {origin_point_counts} 个, 简化后剩余 {new_point_counts} 个, \n"

        cur.execute("select * from laneconnector")  # 根据上表结构建立对应的表结构对象
        laneconnectors = cur.fetchall()
        origin_point_counts = 0
        new_point_counts = 0
        for laneconnector in laneconnectors:
            lane = laneconnector
            points = json.loads(lane["centerLinePointsJson"])
            if not points:
                continue

            points = points['data']
            center_points = [[point['x'], point['y'], point['z']] for point in points]
            new_point_indexs = get_new_point_indexs(center_points, default_angle)

            for key in ['centerLinePointsJson', 'leftBreakPointsJson', 'rightBreakPointsJson']:
                points = json.loads(lane[key])['data']
                new_points = [point for index, point in enumerate(points) if index in new_point_indexs]
                sql = f'update laneconnector set {key}=:who where connID={laneconnector["connID"]} and startLaneid={laneconnector["StartLaneID"]} and endlaneid={laneconnector["EndLaneID"]}'
                cur.execute(sql, {"who": json.dumps({"data": new_points})})

            origin_point_counts += len(points)
            new_point_counts += len(new_points)

        message += f"原连接段中可查询的点位共 {origin_point_counts} 个, 简化后剩余 {new_point_counts} 个 \n"

        conn.commit()  # 保存提交，确保数据保存成功
        conn.close()  # 关闭与数据库的连接

        return message




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytessng",
    version="0.3.11",
    author="yang",
    author_email="17315487709@163.com",
    description="tessng with python3.6",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="",
    install_requires=[
        "attrs==22.2.0",
        "certifi==2021.5.30",
        "charset-normalizer==2.0.12",
        "click==8.0.4",
        "click-plugins==1.1.1",
        "cligj==0.7.2",
        "colorama==0.4.5",
        "commonroad-io==2020.2",
        "cycler==0.11.0",
        "dbfread==2.0.7",
        "decorator==4.4.2",
        "Fiona==1.8.22",
        "geopandas==0.9.0",
        "idna==3.6",
        "importlib-metadata==4.8.3",
        "importlib-resources==5.4.0",
        "kafka-python==2.0.2",
        "kiwisolver==1.3.1",
        "lxml==4.9.4",
        "matplotlib==3.3.4",
        "MouseInfo==0.1.3",
        "munch==4.0.0",
        "networkx==2.5.1",
        "numpy==1.19.5",
        "osmnx==1.1.1",
        "pandas==1.1.5",
        "Pillow==8.4.0",
        "PyAutoGUI==0.9.54",
        "PyGetWindow==0.0.9",
        "PyMsgBox==1.0.9",
        "pyparsing==3.0.7",
        "pyperclip==1.8.2",
        "pyproj==3.0.1",
        "PyRect==0.2.0",
        "PyScreeze==0.1.30",
        "pyshp==2.3.1",
        "PySide2==5.15.2.1",
        "python-dateutil==2.8.2",
        "pytweening==1.0.7",
        "pytz==2023.3.post1",
        "requests==2.27.1",
        "Rtree==0.9.7",
        "scipy==1.5.4",
        "Shapely==1.8.5.post1",
        "shiboken2==5.15.2.1",
        "six==1.16.0",
        "tqdm==4.64.1",
        "typing_extensions==4.1.1",
        "urllib3==1.26.18",
        "wincertstore==0.2",
        "xlrd==1.2.0",
        "zipp==3.6.0",
    ],
    python_requires='>=3.6, <3.7',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=setuptools.find_packages(),
    zip_safe=False,
    include_package_data=True,
)
